# -*- coding: utf-8 -*-
# pylint: disable=C0301,C0103,C0116
"""
Yapp module family_phaser.py

Infer phases and segregation indicators from high density SNP data
in pedigrees.
"""
import sys
import logging
from collections import defaultdict
import time
from multiprocessing import Pool, cpu_count, TimeoutError  # noqa pylint: disable=W0622
import bz2
import pickle
import numpy as np
from scipy.stats import binom
import cyvcf2
from . import vcf, gamete, pedigree

logger = logging.getLogger(__name__)


class PhaseError(Exception):
    """Base class for Phase Errors"""


def qerr(n, p, q=0.001):
    """(1-q) Quantile of the binomial(n,p)"""
    return binom(p=p, n=n).isf(q)


def wcsp_phase(args):
    """parallel WCSP function"""
    indiv, p, children_gametes, recpos = args
    wcsp_gam = gamete.Gamete.from_wcsp_solver(p.g, children_gametes, mkpos=recpos)
    return (indiv, wcsp_gam)


def run_segregation_task(args):
    """parallele segregation function"""
    name, chpair_parent, gam, recmap, err = args
    return name, chpair_parent.get_segregation_indicators(gam, recmap, err)


class ChromosomePair:
    """A class to manipulare a pair of chromosomes"""

    def __init__(self, genotype):
        self.g = gamete.Gamete.valid_genotype(genotype)
        self.H = [
            gamete.Gamete.from_genotype(self.g),
            gamete.Gamete.from_genotype(self.g),
        ]
        self._h_pat = 0
        self.si_pat = [(0, 0.5) for _ in range(len(self.g))]
        self.si_mat = [(0, 0.5) for _ in range(len(self.g))]

    @classmethod
    def from_phase_data(
        cls, genotype, paternal_hap, maternal_hap, paternal_si, maternal_si
    ):
        obj = cls(genotype)
        paternal_gamete = gamete.Gamete(paternal_hap)
        obj.update_paternal_gamete(paternal_gamete)
        maternal_gamete = gamete.Gamete(maternal_hap)
        obj.update_maternal_gamete(maternal_gamete)
        assert len(paternal_si) == obj.len
        assert len(maternal_si) == obj.len
        obj.si_pat = np.array(paternal_si, dtype=int)
        obj.si_mat = np.array(maternal_si, dtype=int)
        return obj

    @property
    def len(self):
        return len(self.g)

    @property
    def het_mks(self):
        "Heterozygous markers"
        return self.g == 1

    @property
    def nhet(self):
        return np.sum(self.het_mks)

    @property
    def phased(self):
        """Phased markers"""
        return (self.paternal_gamete.haplotype > -1) & (
            self.maternal_gamete.haplotype > -1
        )  # noqa

    @property
    def resolved(self):
        """Resolved markers = heterozygotes and phased"""
        return self.het_mks & self.phased

    @property
    def nresolved(self):
        return np.sum(self.resolved)

    @property
    def h_pat(self):
        """Index (0,1) of the paternal haplotype"""
        return self._h_pat

    @h_pat.setter
    def h_pat(self, value):
        assert value == 0 or value == 1
        self._h_pat = value

    @property
    def h_mat(self):
        """Index (0,1) of the maternal haplotype"""
        return 1 - self._h_pat

    @h_mat.setter
    def h_mat(self, value):
        self.h_pat = 1 - value

    @property
    def paternal_gamete(self):
        return self.H[self.h_pat]

    @property
    def maternal_gamete(self):
        return self.H[self.h_mat]

    def get_phase_info(self, origin, call=0.999):
        """Return informative phase in parent origin (0:father, 1:
        mother) = segregation indicator truncated at uninformative
        boundaries.
        """
        assert (origin == 0) or (origin == 1)
        segind = self.si_pat if (origin == 0) else self.si_mat
        phase_info = np.full_like(self.g, -1, dtype=np.int8)
        for i, (a, p) in enumerate(segind):
            if self.phased[i] and p > call:
                phase_info[i] = a
        return phase_info

    def get_transmitted_from_segregation(self, si, call=0.999):
        """
        Returns the gamete transmitted from a segregation indicator vector.
        Inform haplotype only if posterior probability > call
        """
        gam = gamete.Gamete.from_genotype(self.g)
        for i, (orig, prob) in enumerate(si):
            assert orig == 1 or orig == 0
            if prob < call:
                continue
            if orig == 0:
                allele = self.h_pat
            elif orig == 1:
                allele = self.h_mat
            else:
                continue
            gam.haplotype[i] = self.H[allele].haplotype[i]
        return gam

    def get_segregation_indicators(self, gam, recmap, err=1e-2):
        """Infer origin of alleles in the gamete gam using a simple HMM.

        Arguments
        ---------
        gam : gamete.Gamete
           the gamete transmitted
        recmap : array of floats
           the recombination rates between adjacent markers
        err : float
           Genotyping error rate

        Transition probabilities are:
        [ [ (1-r)   r   ]
          [   r   (1-r) ] ... ] for r in recmap
          between all marker pairs

        Returns
        -------
        List( tuple( int, float) )
        segregation indicators for all markers with associated probability
        """
        assert type(gam) == gamete.Gamete
        assert len(gam.haplotype) == self.len
        assert len(recmap) == self.len - 1

        # 1. HMM Setup
        transitions = np.array(
            [np.identity(2) * (1 - x) + (1 - np.identity(2)) * x for x in recmap]
        )
        # [ P(gam[m]| S==pat[m]), P(gam[m]| S==mat[m])]
        emissions = np.ones((self.len, 2), dtype=float)
        for m in range(self.len):
            if (
                (gam.haplotype[m] > -1)
                and self.g[m] > -1
                and self.H[0].haplotype[m] > -1
                and self.H[1].haplotype[m] > -1
            ):
                # ignore mendelian errors
                if (gam.haplotype[m] == 0 and self.g[m] == 2) or (
                    gam.haplotype[m] == 1 and self.g[m] == 0
                ):
                    emissions[m, 0] = 1
                    emissions[m, 1] = 1
                else:
                    emissions[m, 0] = (
                        (gam.haplotype[m] == self.paternal_gamete.haplotype[m])
                        and (1 - err)
                        or err
                    )  # noqa
                    emissions[m, 1] = (
                        (gam.haplotype[m] == self.maternal_gamete.haplotype[m])
                        and (1 - err)
                        or err
                    )  # noqa
        # 2. Forward-Backward algorithm
        fwd = np.ones((self.len, 2), dtype=float)  # Forward
        rew = np.ones((self.len, 2), dtype=float)  # Backward
        sca = np.ones(self.len, dtype=float)  # scaling

        # Compute forward probabilities
        fwd[0, 0] = 0.5 * emissions[0, 0]
        fwd[0, 1] = 0.5 * emissions[0, 1]
        sca[0] = 1.0 / (fwd[0, 0] + fwd[0, 1])
        fwd[0,] *= sca[0]
        for m in range(1, self.len):
            fwd[m, 0] = (
                fwd[m - 1, 0] * transitions[m - 1, 0, 0]
                + fwd[m - 1, 1] * transitions[m - 1, 1, 0]
            ) * emissions[m, 0]
            fwd[m, 1] = (
                fwd[m - 1, 0] * transitions[m - 1, 0, 1]
                + fwd[m - 1, 1] * transitions[m - 1, 1, 1]
            ) * emissions[m, 1]
            sca[m] = 1.0 / (fwd[m, 0] + fwd[m, 1])
            fwd[m,] *= sca[m]
        # Compute backward probabilities
        rew[self.len - 1,] /= sca[m - 1]
        for m in range(self.len - 1, 0, -1):
            rew[m - 1, 0] = (
                transitions[m - 1, 0, 0] * emissions[m, 0] * rew[m, 0]
                + transitions[m - 1, 0, 1] * emissions[m, 1] * rew[m, 1]
            )
            rew[m - 1, 1] = (
                transitions[m - 1, 1, 0] * emissions[m, 0] * rew[m, 0]
                + transitions[m - 1, 1, 1] * emissions[m, 1] * rew[m, 1]
            )
            rew[m - 1,] *= sca[m - 1]
        # loglik = -np.sum(np.log(sca))
        # Compute Posterior probabilities
        post_si = fwd * rew
        post_si /= np.sum(post_si, axis=1, keepdims=True)  # required ?

        # 3. Viterbi Algorithm
        # viterbi variables
        delta = np.zeros((self.len, 2), dtype=float)
        psi = np.zeros((self.len, 2), dtype=int)
        soluce = np.empty(self.len, dtype=int)
        # init
        delta[0, 0] = np.log(0.5 * emissions[0, 0])
        delta[0, 1] = np.log(0.5 * emissions[0, 1])
        # recursion
        for m in range(1, self.len):
            val_0 = [
                delta[m - 1, 0] + np.log(transitions[m - 1, 0, 0]),
                delta[m - 1, 1] + np.log(transitions[m - 1, 1, 0]),
            ]
            val_1 = [
                delta[m - 1, 0] + np.log(transitions[m - 1, 0, 1]),
                delta[m - 1, 1] + np.log(transitions[m - 1, 1, 1]),
            ]
            psi[m, 0] = np.argmax(val_0)
            psi[m, 1] = np.argmax(val_1)
            delta[m, 0] = val_0[psi[m, 0]] + np.log(emissions[m, 0])
            delta[m, 1] = val_1[psi[m, 1]] + np.log(emissions[m, 1])
        # termination / backtracking
        soluce[-1] = np.argmax(delta[-1,])
        for m in range(self.len - 1, 0, -1):
            soluce[m - 1] = psi[m, soluce[m]]
        result = [(x, post_si[i, x]) for i, x in enumerate(soluce)]
        return result

    def update_paternal_gamete(self, prop_gam):
        return self.update_gamete(prop_gam, 0)

    def update_maternal_gamete(self, prop_gam):
        return self.update_gamete(prop_gam, 1)

    def update_unknown_gamete(self, prop_gam, err, q=1e-3):
        """Update the phase when the origin of the prop_gam is not known.
        Tries to update paternal and maternal and keeps the one that leads
        to the fewest # of mismatches.

        """
        misses_pat, new_gam_p = gamete.Gamete.combine(self.paternal_gamete, prop_gam)
        misses_mat, new_gam_m = gamete.Gamete.combine(self.maternal_gamete, prop_gam)
        if min([len(misses_pat), len(misses_mat)]) > qerr(self.nhet, err, q):
            logger.debug(
                f"[update_unknown_gamete] "
                f"{len(misses_pat)} "
                f"{len(misses_mat)} "
                f"{self.nhet}"
            )
            raise PhaseError("Could not resolve origin of proposed gamete")
        if len(misses_mat) < len(misses_pat):
            return self.update_maternal_gamete(prop_gam)
        else:
            return self.update_paternal_gamete(prop_gam)

    def update_gamete(self, prop_gam, parent):
        """Update the gamete of a parent with new_gam.
        Arguments
        ---------
        - prop_gam : Gamete object
           the proposed gamete information
        - parent : int
           0 : update paternal, 1: update maternal
        Returns
        -------
        - [ int, int]
          number of mismatches found in [paternal, maternal] gametes
        """
        nmiss = [0, 0]
        if parent == 0:
            origin = self.h_pat
        elif parent == 1:
            origin = self.h_mat
        else:
            raise ValueError(f"{type(self)}:[update_gamete]: parent must be 0 or 1")
        misses_origin, new_gam = gamete.Gamete.combine(self.H[origin], prop_gam)
        nmiss[origin] = len(misses_origin)
        self.H[origin] = new_gam
        prop_gam_o = gamete.Gamete.complement(self.H[origin], self.g)
        misses_other, new_gam = gamete.Gamete.combine(self.H[1 - origin], prop_gam_o)
        nmiss[1 - origin] = len(misses_other)
        self.H[1 - origin] = new_gam
        return nmiss


class Phaser:
    def __init__(
        self, vcf_file, ped_file, out_prfx, region=None, focalID=None, err=1e-3, rho=1
    ):
        self.prefix = out_prfx
        self.err = err
        self.rho = rho
        if region is not None:
            self.prefix += "_" + region
        if focalID is not None:
            self.prefix += "_" + focalID
        # individuals
        ped = pedigree.Pedigree.from_fam_file(ped_file)
        ped.del_unrelated()
        if focalID is not None:
            ped = ped.get_family(focalID)
        self.ped_file_name = ped_file
        pedindivs = [x.indiv for x in ped]
        # genotype data
        self.data = vcf.vcf2zarr(
            vcf_file, output_prefix=self.prefix, reg=region, samples=pedindivs
        )
        self.vcf_out_file_name = f"{self.prefix}_phased.vcf.gz"
        self.vcf_file_name = f"{out_prfx}.vcf.gz"
        genosmp = defaultdict(lambda: False)
        for g in self.genotyped_samples:
            genosmp[g] = True
        self.ignored_indivs = []
        for indiv in pedindivs:
            if not genosmp[indiv]:
                rm_node = ped.del_indiv(indiv)
                logger.debug(f"Ignoring {indiv} as it is not genotyped")
                self.ignored_indivs.append(rm_node)
        self.pedigree = ped
        self.relations = self.pedigree.to_tuples()
        # Run parameters
        self.timeout_delay = 30

    @classmethod
    def from_prefix(cls, prfx, region=None, focalID=None, err=1e-3, rho=1):
        """
        Create a phaser object from files with the same prefix :
        prfx.vcf.gz and prfx.fam

        Arguments
        ---------

        prfx : str
           Prefix of input files
        **kwargs : arguments
           Optional arguments to pass to yapp.vcf.vcf2fph
        """
        vcf_file = f"{prfx}.vcf.gz"
        fam_file = f"{prfx}.fam"
        obj = cls(vcf_file, fam_file, prfx, region=region, focalID=focalID, err=err)
        print(obj.data.tree())
        return obj

    @property
    def regions(self):
        return self.data["regions"]

    @property
    def genotyped_samples(self):
        return list(self.data["samples"])

    @property
    def pedigree_samples(self):
        return [x for x in self.pedigree.nodes]

    @property
    def families(self):
        return self.pedigree.families

    def recmap(self):
        """Computes a recombination map assuming a rate of recrate (in cM/Mb)
        along regions"""
        res = {}
        for reg in self.regions:
            pos = self.data["variants"][reg]["POS"]
            distances = np.array(pos[1:] - pos[:-1])
            distances[distances < 100] = 100
            res[reg] = distances * self.rho * 1e-8
        return res

    def write_phased_vcf(self):
        """Write phase information in a VCF file"""
        pedindivs = [x.indiv for x in self.pedigree]
        vcf_tmpl = cyvcf2.VCF(self.vcf_file_name, lazy=True, samples=pedindivs)
        w = cyvcf2.Writer(self.vcf_out_file_name, vcf_tmpl)
        snp_mapping = {}
        for s in vcf_tmpl:
            snp_mapping[s.ID] = s
        nidx = np.array([vcf_tmpl.samples.index(n.indiv) for n in self.pedigree])
        for reg in self.regions:
            try:
                snps = self.data["variants"][reg]["ID"]
            except KeyError:
                continue
            variants = [snp_mapping[x] for x in snps]
            ph = np.array(self.data["phases"][reg]["gametes"])
            sel = np.zeros((ph.shape[0], ph.shape[2]), dtype=bool)
            sel = np.where(ph[:, 0, :] > -1, True, False)
            for i, v in enumerate(variants):
                ph_genotypes = v.genotypes[:]
                loc_nidx = nidx[sel[nidx, i]]
                for idx in loc_nidx:
                    ph_genotypes[idx] = [ph[idx, 0, i], ph[idx, 1, i], 1]
                v.genotypes = [list(x) for x in ph_genotypes]
                w.write_record(v)

    ## Make the class pickable even in case of pedigree loops
    def __getstate__(self):
        ## deal with unpickable pedigree
        state = self.__dict__.copy()
        del state["pedigree"]
        return state

    def __setstate__(self, state):
        ## recover pedigree after pickling
        self.__dict__.update(state)
        self.pedigree = pedigree.Pedigree.from_tuples(self.relations)

    def save(self, fname=None):
        """Save Phaser internal state to file"""
        if fname is None:
            fname = f"{self.prefix}_yapp.db"
        dbfile = bz2.BZ2File(fname, "wb")
        pickle.dump(self, dbfile)
        dbfile.close()

    @classmethod
    def load(cls, fname):
        with bz2.BZ2File(fname, "rb") as f:
            obj = pickle.load(f)
        obj.__class__ = cls
        return obj

    def get_genotypes(self, region):
        data = {}
        genotypes = np.array(self.data["genotypes"][region])
        for i, name in enumerate(self.data["samples"]):
            data[name] = genotypes[i,]
        return data

    def get_mendelian_genotypes(self, region):
        genotypes = self.get_genotypes(region)
        for node in self.pedigree:
            pg = genotypes[node.indiv]
            for child in node.children:
                cg = genotypes[child.indiv]
                errors = ((pg == 2) & (cg == 0)) | ((pg == 0) & (cg == 2))
                pg[errors] = -1
                cg[errors] = -1
        return genotypes

    def run(self, ncpu=cpu_count()):
        self.data.create_group("phases")
        with open(self.prefix + "_yapp_phasestats.txt", "w") as fstat:
            print(
                "Individual Sex Generation Nparents \
                Nchildren region het resolved",
                file=fstat,
            )
        for reg in self.regions:
            logger.info(f"Working on region {reg}")
            try:
                phases, ignored = self.phase_samples_from_genotypes(reg, ncpu=ncpu)
            except KeyError:
                logger.info(f"No data on region {reg}")
                continue
            nhet = 0
            nresolved = 0
            for p in phases.values():
                nhet += p.nhet
                nresolved += p.nresolved
            logger.info(
                f"[{reg}] Resolved {nresolved} out of {nhet} phases : "
                f"{100*nresolved/nhet:.1f}%"
            )
            sys.stdout.flush()
            phases = self.phase_samples_from_segregations(
                reg, ncpu=ncpu, phases=phases, ignore_child=ignored
            )
            nhet = 0
            nresolved = 0
            for p in phases.values():
                nhet += p.nhet
                nresolved += p.nresolved
            logger.info(
                f"[{reg}] Resolved {nresolved} out of {nhet} phases : "
                f"{100*nresolved/nhet:.1f}%"
            )
            # Final estimation of segregation indicators
            # once phases are inferred
            _ = self.compute_segregations(reg, phases=phases)

            logger.info("Storing results")
            with open(self.prefix + "_yapp_phasestats.txt", "a") as fstat:
                for node in self.pedigree:
                    p = phases[node.indiv]
                    print(
                        f"{node.indiv}  "
                        f"{node.sex} "
                        f"{node.gen} "
                        f"{(node.father!=None)+(node.mother!=None)} "
                        f"{len(node.children)} "
                        f"{reg} "
                        f"{p.nhet} "
                        f"{p.nresolved}",
                        file=fstat,
                    )

            nind, nsnp = self.data["genotypes"][reg].shape
            ph_g = self.data["phases"].create_group(reg)
            gametes = np.full((nind, 2, nsnp), dtype="int8", fill_value=-2)
            segregations = np.full((nind, 2, nsnp), dtype="int8", fill_value=0)
            seg_probs = np.full(shape=(nind, 2, nsnp), dtype="float", fill_value=0.5)
            for i, name in enumerate(self.data["samples"]):
                try:
                    chpair = phases[name]
                except KeyError:
                    continue
                gametes[i, 0] = chpair.paternal_gamete.haplotype
                gametes[i, 1] = chpair.maternal_gamete.haplotype
                segregations[i, 0] = np.array([x[0] for x in chpair.si_pat])
                segregations[i, 1] = np.array([x[0] for x in chpair.si_mat])
                seg_probs[i, 0] = np.array([x[1] for x in chpair.si_pat])
                seg_probs[i, 1] = np.array([x[1] for x in chpair.si_mat])
            ph_g["gametes"] = gametes
            ph_g["segregations"] = segregations
            ph_g["seg_probs"] = seg_probs
        print(self.data.tree())

    def compute_segregations(self, region, phases=None):
        """Compute segregations for all individuals in the pedigree"""
        logger.info(f"Computing segregations on region {region}")
        recmaps = self.recmap()
        if phases is None:
            chrom_pairs = self.phase_samples_from_genotypes(region)
        else:
            chrom_pairs = phases
        recmap = recmaps[region]
        pat_seg_tasks = []
        mat_seg_tasks = []
        for node in self.pedigree:
            chpair = chrom_pairs[node.indiv]
            if node.father:
                pat_seg_tasks.append(
                    (
                        node.indiv,
                        chrom_pairs[node.father.indiv],
                        chpair.paternal_gamete,
                        recmap,
                        self.err,
                    )
                )
            if node.mother:
                mat_seg_tasks.append(
                    (
                        node.indiv,
                        chrom_pairs[node.mother.indiv],
                        chpair.maternal_gamete,
                        recmap,
                        self.err,
                    )
                )
        with Pool(cpu_count()) as workers:
            for indiv, segind in workers.imap_unordered(
                run_segregation_task, pat_seg_tasks
            ):
                chpair = chrom_pairs[node.indiv]
                chpair.si_pat = segind
            for indiv, segind in workers.imap_unordered(
                run_segregation_task, mat_seg_tasks
            ):
                chpair = chrom_pairs[node.indiv]
                chpair.si_mat = segind
        return phases

    def phase_samples_from_segregations(
        self,
        region,
        ncpu=1,
        phases=None,
        ignore_child=defaultdict(lambda: False),  # noqa
    ):
        """Infer segregation indicators in the pedigree.

        Arguments
        ---------
        recrate : recombination rate in cM/Mb

        Returns
        -------
        update phases
        """
        logger.info("Phasing samples from segregations")
        recmaps = self.recmap()
        if phases is None:
            chrom_pairs, ignore_child = self.phase_samples_from_genotypes(region)
        else:
            chrom_pairs = phases

        recmap = recmaps[region]
        recpos = np.cumsum([0] + list(recmap)) * 100
        logger.info("\t1. Infer Segregation Indicators")
        pat_seg_tasks = []
        mat_seg_tasks = []
        for node in self.pedigree:
            if ignore_child[node.indiv]:
                logger.debug(f"Ignoring {node.indiv} as requested")
                continue
            chpair = chrom_pairs[node.indiv]
            if node.father:
                pat_seg_tasks.append(
                    (
                        node.indiv,
                        chrom_pairs[node.father.indiv],
                        chpair.paternal_gamete,
                        recmap,
                        self.err,
                    )
                )
            if node.mother:
                mat_seg_tasks.append(
                    (
                        node.indiv,
                        chrom_pairs[node.mother.indiv],
                        chpair.maternal_gamete,
                        recmap,
                        self.err,
                    )
                )

        nmm = 0
        with Pool(cpu_count()) as workers:
            for indiv, segind in workers.imap_unordered(
                run_segregation_task, pat_seg_tasks
            ):
                node = self.pedigree.nodes[indiv]
                chpair_p = chrom_pairs[node.father.indiv]
                chpair = chrom_pairs[node.indiv]
                chpair.si_pat = segind
                maxprob = max([x[1] for x in segind])
                if maxprob < 0.999:
                    logger.debug(
                        f"Ignoring {node.indiv} "
                        f"as segregations cannot be resolved confidently"
                    )
                    ignore_child[node.indiv] = True
                else:
                    old_gam = chpair.paternal_gamete
                    new_gam = chpair_p.get_transmitted_from_segregation(chpair.si_pat)
                    nmiss = chpair.update_paternal_gamete(new_gam)
                    if (nmiss[0] + nmiss[1]) > qerr(
                        chpair.nhet * 2,
                        self.err,
                        q=1e-3 / (len(pat_seg_tasks) + len(mat_seg_tasks)),
                    ):  # noqa
                        logger.warning(
                            f"[from_seg] {node.father.indiv}[pat] -> {node.indiv} :"  # noqa
                            f"{50*(nmiss[0]+nmiss[1])/chpair.nhet:.1g} % mismatch"
                        )  # noqa
                        ignore_child[node.indiv] = True
                        chpair.update_paternal_gamete(old_gam)
                        nmm += 1
            for indiv, segind in workers.imap_unordered(
                run_segregation_task, mat_seg_tasks
            ):
                node = self.pedigree.nodes[indiv]
                chpair_m = chrom_pairs[node.mother.indiv]
                chpair = chrom_pairs[node.indiv]
                chpair.si_mat = segind
                maxprob = max([x[1] for x in segind])
                if maxprob < 0.999:
                    logger.debug(
                        f"Ignoring {node.indiv} "
                        f"as segregations cannot be resolved confidently"
                    )
                    ignore_child[node.indiv] = True
                else:
                    old_gam = chpair.maternal_gamete
                    new_gam = chpair_m.get_transmitted_from_segregation(chpair.si_mat)
                    nmiss = chpair.update_maternal_gamete(new_gam)
                    if (nmiss[0] + nmiss[1]) > qerr(
                        chpair.nhet * 2,
                        self.err,
                        q=1e-3 / (len(pat_seg_tasks) + len(mat_seg_tasks)),
                    ):  # noqa
                        logger.warning(
                            f"[from_seg] {node.mother.indiv}[mat] -> {node.indiv} :"  # noqa
                            f"{50*(nmiss[0]+nmiss[1])/chpair.nhet:.1g} % mismatch"
                        )  # noqa
                        ignore_child[node.indiv] = True
                        chpair.update_maternal_gamete(old_gam)
                        nmm += 1
        logger.info(
            f"\t\tFound {nmm} mismatches out of "
            f"{len(pat_seg_tasks)+len(mat_seg_tasks)} tasks"
        )
        logger.info("\t2. Update parental phases")
        wcsp_tasks = []
        for node in self.pedigree:
            p = chrom_pairs[node.indiv]

            children_gametes = {}
            for child in node.children:
                if ignore_child[child.indiv]:
                    logger.debug(f"Ignoring {child.indiv} as requested")
                    continue
                chpair = chrom_pairs[child.indiv]
                if child.father is node:
                    children_gametes[child.indiv] = chpair.paternal_gamete
                elif child.mother is node:
                    children_gametes[child.indiv] = chpair.maternal_gamete
            if len(children_gametes) > 0:
                wcsp_tasks.append((node.indiv, p, children_gametes, recpos))

        logger.info(f"\t\t Phasing {len(wcsp_tasks)} parents with  WCSP")
        remaining_guys = [x[0] for x in wcsp_tasks]
        with Pool(ncpu) as workers:
            results = workers.imap(wcsp_phase, wcsp_tasks)
            ntimeout = 0
            while True:
                try:
                    indiv, new_gam = results.next(timeout=self.timeout_delay)
                    p = chrom_pairs[indiv]
                    try:
                        p.update_unknown_gamete(new_gam, self.err, 1e-2)
                    except PhaseError:
                        logger.debug(
                            f"Inconsistency in phase data for individual"
                            f"{indiv} on region {region}. Will try to fix it."
                        )
                        si = p.get_segregation_indicators(new_gam, recmap, err=self.err)
                        si_gam = gamete.Gamete.from_offspring_segregation(
                            p.g, new_gam, [x[0] for x in si]
                        )
                        try:
                            chrom_pairs[indiv].update_unknown_gamete(
                                si_gam, self.err, 1e-3
                            )
                        except PhaseError:
                            logger.warning(
                                f"[from_seg] Unresolved Phasing inconsistency "
                                f"for individual {indiv} on region {region}. "
                            )
                        else:
                            logger.debug("Fixed it.")
                    remaining_guys.remove(indiv)
                except StopIteration:
                    break
                except TimeoutError:
                    if ntimeout > 5:
                        logger.warning(
                            f"[WCSP] Max delay attained :"
                            f"aborting {len(remaining_guys)} remaining tasks"
                        )
                        for guy in remaining_guys:
                            logger.info(f"WCSP : {guy} not analyzed")
                        break
                    else:
                        logger.warning(
                            f"[WCSP] Result time out : {len(remaining_guys)}"
                            f"tasks remaining, will wait a bit more ..."
                        )
                        ntimeout += 1
        for node in self.pedigree:
            p = chrom_pairs[node.indiv]
            logger.debug(
                f"[from_seg] {node.indiv} -- [ "
                f"sex:{node.sex} "
                f"gen:{node.gen} "
                f"par:{(node.father is not None)+(node.mother is not None)} "  # noqa
                f"off:{len(node.children)} "
                f"res: {p.nresolved}/{p.nhet}]"
            )

        return chrom_pairs

    def phase_samples_from_genotypes(self, region, ncpu=1, phases=None):
        """Reconstruct paternal and maternal gametes in the pedigree
        based on observed genotypes.
        Returns
        -------
        Dict[ str, Dict[ str, ChromosomePair ]]
            regions as keys, dict( name : ChromosomePair) as value
        """
        logger.info("Phasing samples from genotypes")
        logger.info("\t1. Loading Genotypes")
        genotypes = self.get_mendelian_genotypes(region)
        recmaps = self.recmap()
        recmap = recmaps[region]
        recpos = np.cumsum([0] + list(recmap)) * 100
        if phases is None:
            chrom_pairs = {}
            for node in self.pedigree:
                chrom_pairs[node.indiv] = ChromosomePair(genotypes[node.indiv])
        else:
            chrom_pairs = phases
        ignore_child = defaultdict(lambda: False)
        wcsp_tasks = []
        logger.info("\t2. Constructing gametes")
        for node in self.pedigree:
            name = node.indiv
            p = chrom_pairs[name]

            if node.father is not None:
                geno_p = genotypes[node.father.indiv]
                gam_p = gamete.Gamete.from_genotype(geno_p)
                nmiss = p.update_paternal_gamete(gam_p)
                if (nmiss[0] + nmiss[1]) > qerr(p.nhet * 2, self.err):
                    logger.warning(
                        f"[from_geno] {node.father.indiv}[pat] -> {node.indiv}:"  # noqa
                        f"{50*(nmiss[0]+nmiss[1])/p.nhet:.1g}% mismatches"
                    )
                elif nmiss[0] + nmiss[1] > 0:
                    logger.debug(
                        f"[from_geno] {node.father.indiv}[pat] -> {node.indiv}:"  # noqa
                        f"{50*(nmiss[0]+nmiss[1])/p.nhet:.1g}% mismatches"
                    )
            if node.mother is not None:
                geno_m = genotypes[node.mother.indiv]
                gam_m = gamete.Gamete.from_genotype(geno_m)
                nmiss = p.update_maternal_gamete(gam_m)
                if (nmiss[0] + nmiss[1]) > qerr(p.nhet * 2, self.err):
                    logger.warning(
                        f"[from_geno] {node.mother.indiv}[mat] -> {node.indiv}:"  # noqa
                        f"{50*(nmiss[0]+nmiss[1])/p.nhet:.1g}% mismatches"
                    )
                elif (nmiss[0] + nmiss[1]) > 0:
                    logger.debug(
                        f"[from_geno] {node.mother.indiv}[pat] -> {node.indiv}:"  # noqa
                        f"{50*(nmiss[0]+nmiss[1])/p.nhet:.1g}% mismatches"
                    )

            children_gametes = {}
            for child in node.children:
                geno_off = genotypes[child.indiv]
                if child.father is node:
                    if child.mother is not None:
                        geno_other = genotypes[child.mother.indiv]
                    else:
                        geno_other = None
                elif child.mother is node:
                    if child.father is not None:
                        geno_other = genotypes[child.father.indiv]
                    else:
                        geno_other = None
                gam_off = gamete.Gamete.from_offspring_genotype(
                    geno_off, other_geno=geno_other
                )
                children_gametes[child.indiv] = gam_off
            if len(node.children) > 0:
                wcsp_tasks.append((node.indiv, p, children_gametes, recpos))

        logger.info(f"\t\tPhasing {len(wcsp_tasks)} parents with WCSP")
        start_time = time.time()
        with Pool(ncpu) as workers:
            for indiv, new_gam in workers.imap(wcsp_phase, wcsp_tasks):
                p = chrom_pairs[indiv]
                try:
                    # be liberal in calling switch errors
                    p.update_unknown_gamete(new_gam, self.err, 1e-2)
                except PhaseError:
                    logger.debug(
                        f"Inconsistency in phase data "
                        f"for individual {indiv} on region {region}. "
                        f"I Will try to fix it. "
                    )
                    # correct switch errors
                    si = p.get_segregation_indicators(new_gam, recmap, err=self.err)
                    si_gam = gamete.Gamete.from_offspring_segregation(
                        p.g, new_gam, [x[0] for x in si]
                    )
                    try:
                        # admit a bit more inconsistencies to solve gamete
                        chrom_pairs[indiv].update_unknown_gamete(si_gam, self.err, 1e-3)
                    except PhaseError:
                        logger.warning(
                            f"Unresolved phasing inconsistency "
                            f"for individual {indiv} on region {region}. "
                        )
                    else:
                        logger.debug("Fixed it.")
        elapsed_time = time.time() - start_time
        tot_time = elapsed_time * ncpu
        if len(wcsp_tasks) > 0:
            time_per_task = tot_time / len(wcsp_tasks)
            self.timeout_delay = round(time_per_task) + 1
            logger.debug(f"Set timeout delay to {self.timeout_delay} seconds")
        for node in self.pedigree:
            p = chrom_pairs[node.indiv]
            logger.debug(
                f"[from_geno] {node.indiv} -- [ "
                f"sex:{node.sex} "
                f"gen:{node.gen} "
                f"par:{(node.father is not None)+(node.mother is not None)} "  # noqa
                f"off:{len(node.children)} "
                f"res: {p.nresolved}/{p.nhet}]"
            )

        return chrom_pairs, ignore_child


def main(args):
    prfx = args.prfx

    logger.info("Loading Data")
    phaser = Phaser.from_prefix(
        prfx, region=args.reg, focalID=args.focalID, err=args.err, rho=args.rho
    )
    phaser.run(ncpu=args.c)
    logger.info(
        f"Exporting results to : "
        f"{phaser.vcf_out_file_name} and {phaser.prefix}_yapp.db"
    )
    phaser.write_phased_vcf()
    phaser.save()



# -*- coding: utf-8 -*-
"""
Module gamete.py from yapp

This module exports classes and functions to build and manipulate gametes
and their transmissions from genotype data.

This module works with genotype and haplotype data that are expected to be
(castable to) 1-D numpy arrays of integers.

For genotypes valid values (seel Gamete.valid_genotype static method) are:
    - -1 : missing data
    -  0 : homozygote AA
    -  1 : heterozygote (Aa)
    -  2 : homozygote aa

For haplotypes:
    - -2 : unknown allele
    - -1 : unresolved allele
    -  0 : for allele A
    -  1 : for allele a
"""
import logging
import numpy as np
from yapp import wcsp

logger = logging.getLogger(__name__)


class Gamete:
    """A gamete = an haplotype transmitted during meiosis"""

    def __init__(self, hap=None):
        self.haplotype = hap

    @staticmethod
    def valid_genotype(genotype):
        try:
            geno = np.array(genotype, dtype=int)
        except ValueError:
            print("Genotype must be castable to a numpy array of integers")
            raise
        else:
            try:
                assert len(geno.shape) == 1
            except AssertionError:
                raise ValueError("Genotype must be (castable to) a 1-D numpy array.")
        try:
            assert np.all((geno > -2) & (geno < 3))
        except AssertionError:
            raise AssertionError("Genotype values must be in [-1,0,1,2]")
        return geno

    def __str__(self):
        return " ".join([f"{x<0 and ' -' or x:2}" for x in self.haplotype])

    @classmethod
    def from_genotype(cls, genotype):
        """
        Creates a gamete from a genotype.

        Arguments
        ---------
        - genotype : sequence
           A diploid genotype with values

        Returns
        -------
        A Gamete object with haplotype of the same length as genotype.
        Haplotype values are :
        """

        gam = cls()
        geno = cls.valid_genotype(genotype)
        gam.haplotype = np.full_like(geno, -2)
        gam.haplotype[geno == 1] = -1
        gam.haplotype[geno == 0] = 0
        gam.haplotype[geno == 2] = 1
        return gam

    @classmethod
    def from_offspring_genotype(cls, off_geno, other_geno=None):
        """
        Creates the gamete transmitted to an offspring from its
        genotype, possibly using information from its other
        parent's genotype.

        Arguments
        ---------
        - off_geno : genotype
            genotype of the offspring
        - other_geno : genotype
            genotype of the other parent
        """
        gam = cls.from_genotype(off_geno)
        if other_geno is not None:
            gam_other = cls.from_genotype(other_geno)
            try:
                assert len(gam_other.haplotype) == len(gam.haplotype)
            except AssertionError:
                print("Genotypes have different lengths")
                raise
            new_info_mk = (gam.haplotype == -1) & (gam_other.haplotype > -1)
            gam.haplotype[new_info_mk] = 1 - gam_other.haplotype[new_info_mk]
        return gam

    @classmethod
    def from_offspring_segregation(cls, par_geno, off_gam, off_seg):
        """
        Creates an inferred gamete for the parent given the gamete transmitted
        to an offspring and the segregation indicator of the meiosis.

        Arguments
        ---------
        - off_gam : gamete
           gamete received by the offspring
        - off_seg : array of int
           Segregation indicator of the grand-parental origin of the gamete
        """
        gam = cls.from_genotype(par_geno)
        # gam=cls()
        # gam.haplotype = np.full_like(off_gam.haplotype,-2)
        for i, a in enumerate(off_gam.haplotype):
            if a > -1 and off_seg[i] > -1:
                if par_geno[i] == 1:
                    gam.haplotype[i] = a if off_seg[i] == 1 else (1 - a)
                elif par_geno[i] < 0:
                    gam.haplotype[i] = a
        return gam

    @classmethod
    def from_wcsp_solver(cls, par_geno, dict_child_gam, mkpos=None):
        """Infer the gamete of the parent from transmitted gametes using a
        Weighted Constraint Satisfaction Problem

        Arguments
        ---------
        - par_geno : array of int
           genotype of the parent
        - dict_child_gam : dictionary of Gamete objects
           collection of gametes with key = offspring names

        """
        pg = Gamete.valid_genotype(par_geno)
        het_mk = np.where(pg == 1)[0]
        hap_data = {}
        new_gam = Gamete.from_genotype(pg)
        for k, v in dict_child_gam.items():
            hap_data[k] = list(dict_child_gam[k].haplotype[het_mk])
        if mkpos is not None:
            mkpos = mkpos[het_mk]
        phase_data = wcsp.PhaseData(hap_data, mkpos)
        if len(phase_data.info_mk) > 0:
            resolved_mk = [het_mk[i] for i in phase_data.info_mk]
            S = wcsp.PhaseSolver(
                phase_data.info_mk, phase_data.info_pairs, phase_data.recombination
            )
            S.add_constraints()
            try:
                par_phase = S.solve()
                new_gam.haplotype[resolved_mk] = par_phase
            except RuntimeError:
                logger.warning("WCSP solver failure")
        return new_gam

    @classmethod
    def combine(cls, first, second):
        """Combine two gametes into a new one
        Argument
        --------
        first, second : Gametes
            Gametes to combine
        Returns
        -------
        ( int, Gamete )

            Number of mismatches and resulting gamete. Any mismatch is
        resolved as -1.
        """
        assert type(first) == Gamete
        assert type(second) == Gamete
        assert len(first.haplotype) == len(second.haplotype)
        gam = cls()
        gam.haplotype = np.full_like(first.haplotype, -1)
        misses = []
        for i, alleles in enumerate(zip(first.haplotype, second.haplotype)):
            if alleles[0] < 0 and alleles[1] < 0:
                continue
            elif alleles[0] < 0:
                gam.haplotype[i] = alleles[1]
            elif alleles[1] < 0:
                gam.haplotype[i] = alleles[0]
            else:
                if alleles[0] != alleles[1]:
                    misses.append(i)
                    gam.haplotype[i] = -1
                else:
                    gam.haplotype[i] = alleles[0]
        return misses, gam

    def add(self, other):
        """Add two gametes to form a genotype

        Argument
        --------
        other : object of type Gamete
           the other gamete to add
        Returns
        -------
        numpy 1D array of integers : the genotype obtained
        """
        assert type(other) == Gamete
        assert len(self.haplotype) == len(other.haplotype)
        genotype = np.full_like(self.haplotype, -1)
        for i, alleles in enumerate(zip(self.haplotype, other.haplotype)):
            if alleles[0] < 0 or alleles[1] < 0:
                continue
            g = alleles[0] + alleles[1]
            assert g < 3
            genotype[i] = g
        return g

    @classmethod
    def complement(cls, gam, genotype):
        """Return the complementary gamete needed to form the given genotype.
        In case of incompatibility set both gametes to -1.
        """
        assert type(gam) == Gamete
        geno = Gamete.valid_genotype(genotype)
        newgam = cls()
        newgam.haplotype = np.full_like(gam.haplotype, -1)
        for i, (a, g) in enumerate(zip(gam.haplotype, geno)):
            if g < 0 or a < 0:
                continue
            if g == 1:
                newgam.haplotype[i] = g - a
            else:  # 0 -> (0,0) or 2->(1,1)
                try:
                    assert ((g == 0) and (a == 0)) or ((g == 2) and (a == 1))
                except AssertionError:
                    gam.haplotype[i] = -1
                else:
                    newgam.haplotype[i] = a
        return newgam



# -*- coding: utf-8 -*-
"""
Yapp module origins.py

Trace ancestral origins down a pedigree.
"""
import logging
from collections import defaultdict
import numpy as np
from zarr.errors import ContainsGroupError
from numba import njit, prange
from . import family_phaser

logger = logging.getLogger(__name__)


@njit(parallel=True)
def compute_grm(nindiv, nsnp, origins, outgrm):
    # pylint: disable=not-an-iterable
    for i in range(nindiv):
        for j in prange(i, nindiv):
            val = 0
            for ell in range(nsnp):
                val += (
                    (origins[i, 0, ell] == origins[j, 0, ell])
                    + (origins[i, 1, ell] == origins[j, 0, ell])
                    + (origins[i, 0, ell] == origins[j, 1, ell])
                    + (origins[i, 1, ell] == origins[j, 1, ell])
                )
            outgrm[i, j] += val
            # outgrm[i, j] += (origins[i, 0, ell] == origins[j, 0, ell])
            # outgrm[i, j] += (origins[i, 1, ell] == origins[j, 0, ell])
            # outgrm[i, j] += (origins[i, 0, ell] == origins[j, 1, ell])
            # outgrm[i, j] += (origins[i, 1, ell] == origins[j, 1, ell])


class OriginTracer:
    """Class Implementing routines to trace origins down a pedigree"""

    def __init__(self):
        self.norigins = 0
        self.origins = defaultdict(self._new_origin)

    def _new_origin(self):
        """Return a new origin code, increments number of origins"""
        self.norigins += 1
        return self.norigins

    def run(self, phaser_db):
        """Run an analysis on phaser object"""
        phaser = family_phaser.Phaser.load(phaser_db)
        self.trace_origins(phaser)
        self.ped_grm(phaser)

    def trace_origins(self, phaser):
        logger.info("Tracing Origins down the pedigree")
        smpidx = {}
        for i, name in enumerate(phaser.data["samples"]):
            smpidx[name] = i
        try:
            phaser.data.create_group("linkage")
        except ContainsGroupError:
            logger.warning(
                "Seems Zarr archive already has linkage information: will use this"
            )  # noqa
            logger.warning(
                "If rebuilding linkage info is needed, delete the 'linkage' directory"
            )  # noqa
            return
        for reg in phaser.regions:
            logger.info(f"Working on region {reg}")
            logger.info("Grabbing data")
            segregations = np.array(phaser.data[f"phases/{reg}/segregations"])
            logger.info("Diving in")
            or_z = phaser.data["linkage"].create_group(reg)
            origins = np.zeros_like(segregations, dtype=int)
            for node in phaser.pedigree:
                node_idx = smpidx[node.indiv]
                if node.father is None:
                    orig_0 = self.origins[node.indiv + "_p"]
                    origins[
                        node_idx,
                        0,
                    ] = orig_0
                else:
                    fa_idx = smpidx[node.father.indiv]
                    rows = segregations[
                        node_idx,
                        0,
                    ]
                    cols = np.arange(rows.shape[0])
                    origins[
                        node_idx,
                        0,
                    ] = origins[fa_idx, rows, cols]
                if node.mother is None:
                    orig_1 = self.origins[node.indiv + "_m"]
                    origins[
                        node_idx,
                        1,
                    ] = orig_1
                else:
                    mo_idx = smpidx[node.mother.indiv]
                    rows = segregations[
                        node_idx,
                        1,
                    ]
                    cols = np.arange(rows.shape[0])
                    origins[
                        node_idx,
                        1,
                    ] = origins[mo_idx, rows, cols]
            or_z["origins"] = origins
        logger.info("Origins done")
        # Output global results to txt files
        with open(phaser.prefix + "_yapp_ancestors.txt", "w") as fout:
            print("ancestor\tgamete\tcode", file=fout)
            for k in self.origins:
                name, o = k.split("_")
                print(f"{name}\t{o}\t{self.origins[k]}", file=fout)
        with open(phaser.prefix + "_yapp_ancestral_props.txt", "w") as fout:
            ancprop = defaultdict(lambda: defaultdict(int))
            Atot = 0
            for reg in phaser.regions:
                origins = np.array(phaser.data[f"linkage/{reg}/origins"])
                Atot += 2 * origins.shape[2]
                for node in phaser.pedigree:
                    node_idx = smpidx[node.indiv]
                    for anc_all in origins[
                        node_idx,
                        0,
                    ]:
                        ancprop[node.indiv][anc_all] += 1
                    for anc_all in origins[
                        node_idx,
                        1,
                    ]:
                        ancprop[node.indiv][anc_all] += 1
            print("individual\tcode\tproportion", file=fout)
            for node in phaser.pedigree:
                for anc_all in self.origins.values():
                    print(
                        f"{node.indiv}\t{anc_all}\t{ancprop[node.indiv][anc_all]/Atot}",  # noqa
                        file=fout,
                    )

    def ped_grm(self, phaser):
        """Compute Genomic Relationship Matrix based on ancestral allele
        transmissions in the pedigree.
        """
        logger.info("Computing linkage-based GRM")
        samples = list(phaser.data["samples"])
        N = len(samples)
        Ltot = 0
        GRM = np.zeros((N, N), dtype=float)
        for reg in phaser.regions:
            logger.info(f"\tAccumulate region {reg}")
            origins = np.array(phaser.data[f"linkage/{reg}/origins"])
            L = origins.shape[2]
            compute_grm(N, L, origins, GRM)
            Ltot += L
        GRM /= 2 * Ltot

        logger.info("writing GRM to disk")
        with open(phaser.prefix + "_yapp_pedGRM.txt", "w") as fout:
            print("id1 id2 kinship", file=fout)
            for i in range(N):
                for j in range(i, N):
                    if GRM[i, j] > 0:
                        print(samples[i], samples[j], GRM[i, j], file=fout)
        for i in range(2, N):
            for j in range(1, i):
                GRM[i, j] = GRM[j, i]
        phaser.data["linkage/pedGRM"] = GRM


def main(args):
    prfx = args.prfx
    phaser_db = prfx + ".db"
    tracer = OriginTracer()
    tracer.run(phaser_db)



# -*- coding: utf-8 -*-
"""Module vcf.py from yapp

This module exports functions to parse VCF files into various input
for other modules in yapp or other packages. VCF files are read using cyvcf2.
"""
import sys
import logging
from collections import defaultdict
import numpy as np
import zarr
from cyvcf2 import VCF

logger = logging.getLogger(__name__)

modes_avail = ["genotype", "inbred", "phased", "likelihood"]
default_mode = modes_avail[0]


def geno2int(a1, a2):
    if a1 == -1 or a2 == -1:
        return -1
    else:
        return a1 + a2


def geno2hap(a1, a2):
    if a1 == -1 or a2 == -1 or (a1 != a2):
        return -1
    else:
        return a1


def phred2ln(x):
    """Transforms a PHRED score into a natural logarithm"""
    return -0.1 * np.log(10) * x


def chunk_region(reg, chunksize=-1):
    """Decompose reg into smaller regions
    Parameters
    ----------
    reg : str
       region in bed format ("chr:begin-end").
       begin and end may contain commas.
    chunksize : float
       size of chunks in basepairs

    Returns
    -------
    list of str
        list of sub-regions in bed format.
        If chunksize is <0, returns a list with the original region
    """
    if chunksize < 0:
        return [reg]
    chrom, coords = reg.split(":")
    beg, end = coords.split("-")
    beg = int(beg.replace(",", ""))
    end = int(end.replace(",", ""))
    starts = np.arange(beg, end, step=chunksize, dtype=int)
    ends = [int(x + chunksize - 1) for x in starts]
    ends[-1] = end
    reglist = []
    for i, s in enumerate(starts):
        newreg = chrom + ":" + str(s) + "-" + str(ends[i])
        reglist.append(newreg)
    return reglist


def vcf_chunk_regions(fname, chunksize=-1):
    """Creates regions of chunksize in a VCF file from contig header information
    Parameters
    ----------

    fname : str
        The name of the vcf(.gz) file
    chunksize : int
        Size of chunks

    Returns
    -------
    list of str
        List of regions in bed format. If chunksize is <0, regions are contigs.
    """
    v = VCF(fname, lazy=True)
    regions = [name + ":" + "0-" + str(l) for name, l in zip(v.seqnames, v.seqlens)]
    if chunksize > 0:
        new_regions = []
        for r in regions:
            new_regions += chunk_region(r, chunksize)
        regions = new_regions
    return regions


def vcf2fph(
    fname,
    mode=default_mode,
    samples=None,
    reg=None,
    maf=0.01,
    varids=None,
    genotypes=True,
):
    """Parses a VCF file and returns data arrays usable with
    the fastphase package

    Parameters
    ----------
    fname : str
        The name of the vcf(.gz) file
    mode : str
        One of 'genotype', 'inbred', 'phased' or 'likelihood'.

        In 'genotype' mode, the returning 1D arrays are genotypes
        (0,1,2 and -1 for missing data)

        In 'inbred' mode, the resulting 1D arrays are haplotypes (0,1 and -1).
        Heterozygote calls are treated as missing data.

        In 'phased' mode, two haplotype 1D arrays are returned per
        sample, using phased information in the VCF (eg. 0|1). Any non
        phased genotype is considered missing.

        In 'likelihood' mode, the resulting 2D arrays contain the
        genotype log-likelihoods extracted from the 'PL' field.

    samples : list
        List of samples to extract from full set in file
    reg : str
        query variants in region reg only in bed format ("chr:begin-end").
        VCF file must be indexed.
    maf : double
        Minimum allele frequency to include variants
    varids : list
        List of variants IDs to extract. Only variants found in vcf are
        returned, irrespective of MAF.
    genotypes : bool
        Return Genotype data. For large files this can be quite big.
    Returns
    -------
    dict
        {
         regions : list of regions (str)
         samples : list of samples (str)
         variants : dict {region : list of variants (ID, CHROM, POS, REF, ALT)}
         data : dict {region : dict {
                                  sample ID: data (numpy array)
                                  data type depends on mode. see Parameters.)}
        }

    """

    v = VCF(fname, gts012=True, strict_gt=True, samples=samples, lazy=True)
    smp = v.samples  # might differ if some requested are not found

    if reg is None:
        regions = v.seqnames
    else:
        regions = [reg]

    if varids is None:
        keepvar = defaultdict(lambda: True)
    else:
        maf = 0
        keepvar = defaultdict(lambda: False)
        for s in varids:
            keepvar[s] = True
    variants = {}
    trueregions = []
    variants_summary = {}
    fphdata = {}
    for r in regions:
        snps = []
        for s in v(r):
            if len(s.ALT) > 1 or s.aaf < maf or not keepvar[s.ID]:
                continue
            snps.append(s)
        if len(snps) == 0:
            continue
        variants[r] = sorted(snps, key=lambda x: x.POS)
        variants_summary[r] = [
            (s.ID, s.CHROM, s.POS, s.REF, s.ALT[0]) for s in variants[r]
        ]
        fphdata[r] = {}
        trueregions.append(r)
        if genotypes:
            for i, sid in enumerate(smp):
                if mode == "likelihood":
                    fphdata[r][sid] = phred2ln(
                        np.array(
                            [
                                [
                                    s.gt_phred_ll_homref[i],
                                    s.gt_phred_ll_het[i],
                                    s.gt_phred_ll_homalt[i],
                                ]
                                for s in variants[r]
                            ],
                            dtype=float,
                        )
                    )
                else:
                    geno = [s.genotypes[i] for s in variants[r]]
                    if mode == "genotype":
                        fphdata[r][sid] = np.array(
                            [geno2int(*g[:2]) for g in geno], dtype=int
                        )
                    elif mode == "inbred":
                        fphdata[r][sid] = np.array(
                            [geno2hap(*g[:2]) for g in geno], dtype=int
                        )
                    elif mode == "phased":
                        h1 = []
                        h2 = []
                        for g in geno:
                            if g[2]:
                                h1.append(g[0])
                                h2.append(g[1])
                            else:
                                h1.append(-1)
                                h2.append(-1)
                        fphdata[r][sid + ".h1"] = np.array(h1, dtype=int)
                        fphdata[r][sid + ".h2"] = np.array(h2, dtype=int)
    return {
        "regions": trueregions,
        "samples": smp,
        "variants": variants_summary,
        "data": fphdata,
    }


def vcf2zarr(
    fname,
    output_prefix=None,
    mode=default_mode,
    samples=None,
    reg=None,
    maf=0.01,
    varids=None,
):
    """Parses a VCF file and creates a ZARR file with its contents

    Parameters
    ----------
    fname : str
        The name of the vcf(.gz) file
    output_prefix: str
        The prefix for zarr file name (prefix.zarr)
    mode : str
        One of 'genotype', 'inbred', 'phased' or 'likelihood'.

        In 'genotype' mode, the returning 1D arrays are genotypes
        (0,1,2 and -1 for missing data)

        In 'inbred' mode, the resulting 1D arrays are haplotypes (0,1 and -1).
        Heterozygote calls are treated as missing data.

        In 'phased' mode, two haplotype 1D arrays are returned per
        sample, using phased information in the VCF (eg. 0|1). Any non
        phased genotype is considered missing.

        In 'likelihood' mode, the resulting 2D arrays contain the
        genotype log-likelihoods extracted from the 'PL' field.

    samples : list
        List of samples to extract from full set in file
    reg : str
        query variants in region reg only in bed format
        ("chr:begin-end"). VCF file must be indexed.
    maf : double
        Minimum allele frequency to include variants
    varids : list
        List of variants IDs to extract. Only variants found in vcf
        are returned, irrespective of MAF.
    genotypes : bool
        Return Genotype data. For large files this can be quite big.
    Returns
    -------
    str
       handler to the zarr root group

    """

    v = VCF(fname, gts012=True, strict_gt=True, samples=samples, lazy=True)
    if not output_prefix:
        output_prefix = fname.split(".")[0]
    fzname = output_prefix + ".zarr"
    fz = zarr.open(fzname, "w")
    fz.attrs["archive"] = fzname
    smp = v.samples  # might differ if some requested are not found
    if reg is None:
        _regions = v.seqnames
    else:
        _regions = [reg]

    regions = []
    for r in _regions:
        ns = len([s for s in v(r)])
        if ns > 0:
            regions.append(r)
    if len(regions) < 1:
        raise IOError(f"No SNP found in {fname}")

    logger.info("Creating Archive structure")
    fz["samples"] = zarr.array(smp, dtype="U50")
    fz["regions"] = zarr.array(regions, dtype="U50")
    for i, reg in enumerate(regions):
        fz.create_group(f"variants/{reg}")
    fz.create_group("genotypes")

    if varids is None:
        keepvar = defaultdict(lambda: True)
    else:
        maf = 0
        keepvar = defaultdict(lambda: False)
        for s in varids:
            keepvar[s] = True

    for r in regions:
        logger.info(f"Getting data for region {r}")
        snps = []
        for s in v(r):
            if len(s.ALT) > 1 or s.aaf < maf or not keepvar[s.ID]:
                continue
            snps.append(s)
        if len(snps) == 0:
            continue
        variants = sorted(snps, key=lambda x: x.POS)
        fz[f"variants/{r}/ID"] = np.array([s.ID for s in variants], dtype="U50")
        fz[f"variants/{r}/CHROM"] = np.array([s.CHROM for s in variants], dtype="U50")
        fz[f"variants/{r}/POS"] = np.array([s.POS for s in variants], dtype=np.uint32)
        fz[f"variants/{r}/REF"] = np.array([s.REF for s in variants], dtype="S20")
        fz[f"variants/{r}/ALT"] = np.array([s.ALT[0] for s in variants], dtype="S20")

        data = []
        for i, sid in enumerate(smp):
            sys.stdout.write(f"{sid:<20} {100*(i+1)/len(smp):.2f} %\r")
            sys.stdout.flush()
            if mode == "likelihood":
                gind = phred2ln(
                    np.array(
                        [
                            [
                                s.gt_phred_ll_homref[i],
                                s.gt_phred_ll_het[i],
                                s.gt_phred_ll_homalt[i],
                            ]
                            for s in variants[r]
                        ],
                        dtype=float,
                    )
                )
            else:
                geno = [s.genotypes[i] for s in variants]
                if mode == "genotype":
                    gind = np.array([geno2int(*g[:2]) for g in geno], dtype=np.int8)
                elif mode == "inbred":
                    gind = np.array([geno2hap(*g[:2]) for g in geno], dtype=np.int8)
                elif mode == "phased":
                    gind = np.array(
                        [
                            [g[0] if g[2] else -1 for g in geno],
                            [g[1] if g[2] else -1 for g in geno],
                        ],
                        dtype=np.int8,
                    )
            data.append(gind)
        fz[f"genotypes/{r}"] = zarr.array(data, chunks=False)
    return fz



# -*- coding: utf-8 -*-
"""
Yapp module sperm.py

Infer genotype and phase of a parent from genotyping data of its gametes
"""
import logging
from collections import defaultdict
import numpy as np
from scipy.stats import binom
from . import vcf, gamete, pedigree, family_phaser, MALE, FEMALE  # noqa

logger = logging.getLogger(__name__)


def genotype_from_gametes(gametes, pgeno=0.95, gerr=1e-2):
    """Infer genotype from an array of gametes

    Parameters
    ----------
    gametes : numpy array of int
       The haplotypes of gametes. Rows are gametes, columns are SNPs.
    0 = REF, 1 = ALT, -1 = Missing
    pgeno : float
       Minimum probability to call a genotype

    gerr : float
       Genotyping error rate

    Returns
    -------
    numpy array of int
        genotype of the parent
        0 = REF/REF, 1 =REF/ALT, 2 = ALT/ALT, -1 = Missing

    """
    nchrobs = np.sum(gametes > -1, axis=0)
    nalt = np.sum(gametes == 1, axis=0)
    lik = np.zeros((3, nalt.shape[0]), dtype=float)
    lik[0,] = binom.pmf(nalt, nchrobs, gerr)
    lik[1,] = binom.pmf(nalt, nchrobs, 0.5)
    lik[2,] = binom.pmf(nalt, nchrobs, 1 - gerr)
    lik /= np.sum(lik, axis=0, keepdims=True)
    bestg = np.argmax(lik, axis=0)
    bestp = np.max(lik, axis=0)
    genotype = np.where(bestp > pgeno, bestg, -1)
    return genotype


def recmap(phys_pos, recrate=1):
    distances = np.array(phys_pos[1:] - phys_pos[:-1])
    distances[distances < 100] = 100
    return distances * recrate * 1e-8


def main(args):
    prfx = args.prfx
    err = args.err
    try:
        assert (args.minsp > 0) and (args.minsp < 1)
    except AssertionError:
        logger.error(f"Unvalid argument passed for parameter minsp ({args.minsp})")
        raise
    vcf_file = f"{prfx}.vcf.gz"
    fam_file = f"{prfx}.fam"

    ped = pedigree.Pedigree.from_fam_file(
        fam_file, parent_from_FID=True, default_parent=FEMALE
    )
    for fam in ped.families:
        assert len(fam.founders) == 1
        trgt_i = fam.founders[0].indiv
        oprfx = f"{prfx}_{trgt_i}"
        logger.debug(f"{trgt_i} with {len(fam.non_founders)} gametes")
        logger.info(f"Getting gamete data for individual {trgt_i}")
        gam_data = vcf.vcf2zarr(
            vcf_file,
            output_prefix=oprfx,
            samples=[x.indiv for x in fam.non_founders],
            mode="inbred",
            maf=0,
        )
        gam_data.create_group("parent")
        logger.debug(f"\n {gam_data.tree()}")
        with open(f"{oprfx}.tped", "w") as ftped:
            for reg in gam_data["regions"]:
                logger.info(f"[{trgt_i}] Region {reg}")
                gam_hap = np.array(gam_data[f"genotypes/{reg}"])
                g = genotype_from_gametes(gam_hap, pgeno=args.pgeno, gerr=args.err)
                chrom_pair = family_phaser.ChromosomePair(g)
                children_gametes = dict(
                    zip(gam_data["samples"], [gamete.Gamete(h) for h in gam_hap])
                )

                logger.info("\tPhase from genotypes")
                wcsp_gam = gamete.Gamete.from_wcsp_solver(g, children_gametes)
                logger.debug(f"\n {gam_hap[:, :10]}")
                logger.debug(f"{g[:10]}")
                logger.debug("-" * 10)
                logger.debug(f"WCSP: {wcsp_gam.haplotype[:10]}")
                chrom_pair.update_unknown_gamete(wcsp_gam, err)
                logger.debug(f"h.pat: {chrom_pair.paternal_gamete.haplotype[:10]}")
                logger.debug(f"h.mat: {chrom_pair.maternal_gamete.haplotype[:10]}")

                logger.info("\tPhase from segregations")
                reg_map = recmap(
                    np.array(gam_data[f"variants/{reg}/POS"]), recrate=args.rho
                )
                inconsistencies = [defaultdict(int), defaultdict(int)]
                for sperm in children_gametes.values():
                    si_sperm = chrom_pair.get_segregation_indicators(
                        sperm, recmap=reg_map
                    )
                    # impute paternal and maternal gametes at missing genotypes
                    new_gametes = [
                        gamete.Gamete(chrom_pair.paternal_gamete.haplotype),
                        gamete.Gamete(chrom_pair.maternal_gamete.haplotype),
                    ]
                    for i, geno in enumerate(g):
                        if si_sperm[i][1] > args.minsp:
                            origin = si_sperm[i][0]
                            if geno == -1:
                                if new_gametes[origin].haplotype[i] < 0:
                                    new_gametes[origin].haplotype[i] = sperm.haplotype[
                                        i
                                    ]  # noqa
                                elif (
                                    new_gametes[origin].haplotype[i]
                                    != sperm.haplotype[i]
                                ):  # noqa
                                    inconsistencies[origin][i] += 1
                for locus in inconsistencies[0]:
                    new_gametes[0].haplotype[locus] = -1
                for locus in inconsistencies[1]:
                    new_gametes[1].haplotype[locus] = -1
                logger.debug(f"h.pat: {len(inconsistencies[0])} loci with mismatches")
                logger.debug(f"h.mat: {len(inconsistencies[1])} loci with mismatches")
                chrom_pair.update_paternal_gamete(new_gametes[0])
                chrom_pair.update_maternal_gamete(new_gametes[1])

                logger.debug(f"h.pat: {chrom_pair.paternal_gamete.haplotype[:10]}")
                logger.debug(f"h.mat: {chrom_pair.maternal_gamete.haplotype[:10]}")
                ##
                # save results
                pos = gam_data["variants"][reg]["POS"]
                chrom = gam_data["variants"][reg]["CHROM"]
                rsid = gam_data["variants"][reg]["ID"]
                ref = gam_data["variants"][reg]["REF"]
                alt = gam_data["variants"][reg]["ALT"]

                ph_par = gam_data["parent"].create_group(reg)
                gametes = np.full((2, pos.shape[0]), dtype="int8", fill_value=-2)
                gametes[0] = chrom_pair.paternal_gamete.haplotype
                gametes[1] = chrom_pair.maternal_gamete.haplotype
                ph_par["gametes"] = gametes
                for i, rs in enumerate(rsid):
                    if chrom_pair.paternal_gamete.haplotype[i] == -1:
                        pat_a = b"0"
                    elif chrom_pair.paternal_gamete.haplotype[i] == 0:
                        pat_a = ref[i]
                    else:
                        pat_a = alt[i]
                    if chrom_pair.maternal_gamete.haplotype[i] == -1:
                        mat_a = b"0"
                    elif chrom_pair.maternal_gamete.haplotype[i] == 0:
                        mat_a = ref[i]
                    else:
                        mat_a = alt[i]
                    print(
                        f"{chrom[i]} {rs} 0 {pos[i]} "
                        f"{pat_a.decode()} {mat_a.decode()}",
                        file=ftped,
                    )
    return



# -*- coding: utf-8 -*-
"""
Yapp module lmm.py

Functions to fit linear mixed-models for linkage and LDL analysis.  The
models considered are multivariate gaussian models of the form:

Y = XB + Zu + a + e

- XB are fixed effects terms
- a are the animal random effect (possibly with design matrix if 
  repeated measures) with a ~N(0,G.va)
- u is an additional random effect, typically of ancestral/founder 
  haplotypes, with design matrix Z.

Although some functions can be quite general, the philosophy is to target
specific models used in yapp. This is *not* the most computationaly
efficient module to do general mixed-model analysis.
"""

from functools import partial
import numpy as np
from scipy.optimize import minimize
from scipy.special import expit
from scipy.linalg import block_diag
from scipy.stats import f as fdist
from scipy.stats import chi2


def relik_animal_model(h, Y, X, Gi):
    """
    Computes the restricted likelihood of the animal model:

    Y = XB + a + e

    a ~N( 0, G ve.h/(1-h))
    e ~N( 0, I ve)

    Arguments:
    ----------
    - h : float
        heritability (h in ]0,1[)
    - Y : vector
        Observations
    - Gi : matrix like
        Inverse of G
    - X : matrix like
        Design Matrix of Fixed Effects.
    """
    pass


def relik_Gdiag(h, Y, X, diagG):
    """Computes the restricted likelihood of the animal model when G
    is diagonal

    Y = XB + w + e

    w ~ N(0, diagG ve.h/(1-h))
    e ~ N(0, I ve)

    This is obtained from the full matrix model by diagonalization of
    G through eigen decomposition and projection of the model on the
    PCs.

    V(Y) is diagonal (I + h/(1-h) diagG) and easy to
    invert. So the restricted likelihood can be obtained as :

    C = Xt.Vi.X
    CB = Xt.Vi.Y

    L = P(Y|X,h) = \int P(Y|X,h,ve)P(ve) with P(ve) \propto 1/ve
    logL = cste - 0.5 ( logdet(C) + (N-P) log(Yt.Vi.Y - Bt.C.B) )
    """
    N = Y.shape[0]
    P = X.shape[1]
    Vi = np.diag(1 / (1 + diagG * h / (1 - h)))
    C = X.T @ Vi @ X
    B = np.linalg.solve(C, X.T @ Vi @ Y)
    yPy = Y.T @ Vi @ Y - B.T @ C @ B
    _, ldetC = np.linalg.slogdet(C)
    ldetVi = np.sum(np.log(1 / (1 + diagG * h / (1 - h))))
    return -0.5 * (ldetC - ldetVi + (N - P) * np.log(yPy))


def Wald_Gdiag(h, Y, X, diagG):
    """Computes the Wald statistic of fixed effects of the animal
    model when G is diagonal
    """
    N = Y.shape[0]
    P = X.shape[1]
    Vi = np.diag(1 / (1 + diagG * h / (1 - h)))
    C = X.T @ Vi @ X
    B = np.linalg.solve(C, X.T @ Vi @ Y)
    return (B.T @ C @ B, P - 1, N - P)


def relik_lowrank(h, Y, X, eigVal, eigVec, valmin=1):
    """Computes the restricted likelihood of the animal model when G
    is appromixated by (eigVec.sqrt(lambda))(eigVec.sqrt(lambda))'
    where only a subset of eigenValues are kept.

    In this case, with W=eigVec, L=eigVal,

    Vi \approx I_N - vg.W(1/L + I)^{-1}W'

    C = Xt.Vi.X
    CB = Xt.Vi.Y

    L = P(Y|X,h) = \int P(Y|X,h,ve)P(ve) with P(ve) \propto 1/ve
    logL = cste - 0.5 ( logdet(C) + logdet(V) + N log(Yt.Vi.Y - Bt.C.B) )

    References:
    -----------
    - Mischenko et al. 2007, Efficient implementation of the AI-REML
    iteration for variance component QTL analysis

    """
    N = Y.shape[0]
    vg = h / (1 - h)
    L = eigVal[eigVal > valmin]
    W = eigVec[:, eigVal > valmin]
    K = W.shape[1]
    # print("*** K=", K)
    # print(min(eigVal),max(eigVal))
    ## approximate inverse(V) using low rank matrix (Mischenko et al. 2007)
    LRD = np.diag(1 / (vg + 1 / L))
    Vi = np.eye(N) - vg * W @ LRD @ W.T
    ## then as usual
    C = X.T @ Vi @ X
    B = np.linalg.solve(C, X.T @ Vi @ Y)
    yPy = Y.T @ Vi @ Y - B.T @ C @ B
    _, ldetC = np.linalg.slogdet(C)
    ldetVi = -np.sum(np.log(vg * L + 1))  ## should we approximate this
    ## too ?
    return -0.5 * (ldetC - ldetVi + N * np.log(yPy))


def relik_Gdiag_Z(p, h, Y, X, Z, diagG):
    """Computes the restricted likelihood of the animal model when G
    is diagonal and a second random effect is included with design
    matrix Z.

    Y = XB + Zu + w + e

    w ~ N(0, diagG ve.(1-p).h/(1-h))
    u ~ N(0, Ik ve.p.h/(1-h))
    e ~ N(0, I ve)

    We rewrite the model as:
    Y = XB + Zu + r

    with r ~ N(O,R.ve) and R is diagonal (I + (1-p).h/(1-h) diagG).
    with u ~ N(0,K.ve) and K is diagonal Ik.p.h/(1-h).
    The mixed model equations are:

    C = [ [ XtRiX    ,    XtRiZ ] ,
          [ ZtRiX    , ZtRiZ + Ki]]

    C[B:u]t = [X:Z]t.Ri.Y
    """
    N = Y.shape[0]
    P = X.shape[1]
    K = Z.shape[1]
    Vi = np.diag(1 / (1 + diagG * (1 - p) * h / (1 - h)))
    W = np.hstack([X, Z])
    C = W.T @ Vi @ W
    C[P:, P:] += np.eye(K) * (1 - h) / (p * h)
    B = np.linalg.solve(C, W.T @ Vi @ Y)
    yPy = Y.T @ Vi @ Y - B.T @ C @ B
    _, ldetC = np.linalg.slogdet(C)
    ldetVi = np.sum(np.log(1 / (1 + diagG * (1 - p) * h / (1 - h))))
    return -0.5 * (ldetC - ldetVi + (N - P) * np.log(yPy))


if __name__ == "__main__":
    """Simulate stuff to test funcs"""

    rng = np.random.default_rng()
    Nfam = 10
    Nind = 100
    N = Nfam * Nind
    K = 2 * Nfam
    nsim = 1
    ## pedigree kinship -> does not work for testing linkage effects
    ##G = np.kron(np.eye(Nfam),np.ones((Nind,Nind))*0.25)
    fG = []
    ## create pedGRM-like coefficients with gametic variance
    for ifam in range(Nfam):
        a = rng.normal(loc=0.25, scale=0.05, size=(Nind, Nind))
        fG.append(np.tril(a) + np.tril(a, k=-1).T)
    G = block_diag(*fG)
    print(np.round(G, 2))
    np.fill_diagonal(G, 1)
    dG, U = np.linalg.eigh(G)
    with open("../test/h.txt", "w") as fout:
        print("h hreml success ll se", file=fout)
        for isim in range(nsim):
            # h = rng.random()
            # p = rng.random()
            h = 0.5
            p = 0.00001
            print(f"Simulation {isim+1} with h={h} and p={p}")
            ## polygenic component
            w = rng.multivariate_normal(mean=np.zeros(N), cov=G * (1 - p) * h / (1 - h))
            ## Fixed Effects = intercept
            # beta = np.array([2,0.01])
            # X = np.block([[np.ones(N)],[np.random.randint(0,2,N)]]).T
            X = np.block([[np.ones(N)]]).reshape(N, 1)
            beta = np.array([0])
            ## QTL Effect
            Z = np.zeros((N, 2 * Nfam))
            irow = np.arange(N)
            fam, row_within_fam = divmod(irow, Nind)
            div = np.sum(divmod(Nind, 2))
            icol = 2 * fam + (row_within_fam // div)
            Z[irow, icol] = 1
            ##Z = rng.binomial(1,0.5, (N,2*Nfam))
            u = rng.normal(scale=np.sqrt(p * h / (1 - h)), size=Z.shape[1])
            Zu = Z @ u
            ## Phenotypes
            Ynull = X @ beta + w + np.random.randn(N)
            Y = Ynull + Zu
            ##Y = X@beta + Zu + w + np.random.randn(N)
            ## Null Model
            mylik_null = partial(relik_Gdiag, Y=U.T @ Y, X=U.T @ X, diagG=dG)
            optlik_null = lambda x: -mylik_null(x)
            hmax_null = minimize(
                optlik_null, [0.4], bounds=[(1e-3, 1 - 1e-3)], options={"disp": False}
            )
            print(
                f"*****   H: {hmax_null.x[0]} se {np.sqrt(hmax_null.hess_inv.todense()[0,0])}"
            )
            ## QTL model fixed effects
            ##W = np.hstack([X,Z])
            mylik_qtl = partial(
                relik_Gdiag, Y=U.T @ Y - U.T @ X @ beta, X=U.T @ Z, diagG=dG
            )
            optlik_qtl = lambda x: -mylik_qtl(x)
            hmax_qtl = minimize(
                optlik_qtl, [0.4], bounds=[(1e-3, 1 - 1e-3)], options={"disp": False}
            )
            print(
                f"*****   H: {hmax_qtl.x[0]} se {np.sqrt(hmax_qtl.hess_inv.todense()[0,0])}"
            )
            wald = Wald_Gdiag(hmax_qtl.x[0], Y=U.T @ Y, X=U.T @ Z, diagG=dG)
            print(
                f"Wald statistic : {wald}, p = {fdist.sf(wald[0]/wald[1],wald[1],wald[2])} "
            )
            continue
            ## QTL model random effects
            mylik_qtl_2 = partial(
                relik_Gdiag_Z,
                Y=U.T @ Y - U.T @ X @ beta,
                X=U.T @ X,
                Z=U.T @ Z,
                diagG=dG,
            )
            optlik_qtl_2 = lambda x: -mylik_qtl_2(x[0], x[1])
            hmax_qtl_2 = minimize(
                optlik_qtl_2,
                [0.1, 0.1],
                bounds=[(1e-3, 1 - 1e-3), (1e-3, 1 - 1e-3)],
                options={"disp": False},
            )
            H = hmax_qtl_2.hess_inv.todense()
            print(f"Hessian : {H}")
            print(f"*****  H2: {hmax_qtl_2.x[1]} se {H[1,1]}")
            print(f"***** RHO: {hmax_qtl_2.x[0]} se {H[0,0]}")
            #
            # print(np.round(h,3),
            #       np.round(hmax.x[0],3),
            #       hmax.success,
            #       -hmax.fun,
            #         np.sqrt(hmax.hess_inv.todense()[0,0]),
            #         file=fout)

    # h = 0.5
    # genetic = True
    # if genetic:
    #     ## 1. Nfam families of halfsibs
    #     G = np.kron(np.eye(Nfam),np.ones((Nind,Nind))*0.25)
    #     np.fill_diagonal(G,1)
    #     dG,U = np.linalg.eigh(G)
    #     Zu = rng.multivariate_normal(mean = np.zeros(N),cov=G*h/(1-h))
    # else:
    #     ## 2. random vectors
    #     Z = rng.normal(size=(N,N))
    #     dG,U = np.linalg.eigh(Z@Z.T)
    #     u = rng.normal(size=N,scale=np.sqrt(h/(1-h)))
    #     Zu = Z@u

    # ## Fit REML
    # X = np.block([[np.ones(N)],[np.random.randint(0,2,N)]]).T
    # beta = np.array([2,0.01])
    # Y = X@beta + Zu + np.random.randn(N)
    # hh = np.linspace(0.01,0.9999)
    # with open('../test/h.txt','w') as fout:
    #     print('hh ll.diag',file=fout)
    #     mylik = partial(relik_Gdiag_correct, Y=U.T@Y, X=U.T@X, diagG=dG)
    #     optlik = lambda x: -mylik(x)
    #     hmax = minimize(optlik,[0.4],bounds=[(1e-3,1-1e-3)], options={"disp":True})
    #     print(f" h2 = {hmax.x}, success = {hmax.success}")
    #     print(np.round(hmax.x[0],3), -hmax.fun, file=fout)
    #     for x in hh:
    #         res = relik_Gdiag_correct(x, U.T@Y, U.T@X, dG)
    #         print(np.round(x,2), res,file=fout)



# -*- coding: utf-8 -*-
"""Module pedigree.py from yapp

This module exports classes and functions to read and manipulate
pedigrees.  A pedigree is a directed graph of familial relationships
(father, mother, offspring).

"""
import sys
import warnings
import logging
from . import MALE, FEMALE

logger = logging.getLogger(__name__)


class PedNode:
    """A node in the pedigree that consists of an individual and its
    direct relatives.

        Attributes
        ----------
        indiv : obj
           the individual data. Must be hashable.
        father : obj
           the individual's father node. Must be a PedNode object.
        mother : obj
           the individual's mother node. Must be a PedNode object.
        children : list of obj
           the individual's offspring. Must be PedNode objects.

    """

    def __init__(self, indiv, father=None, mother=None, sex=None):
        self.__indiv = indiv
        self.__father = None
        self.__mother = None
        self.__sex = None
        self.__gen = None
        self.father = father
        self.mother = mother
        self.sex = sex
        self.children = []

    def __str__(self):
        output = (
            f"*** {self.__indiv} ***\n"
            f"Father   : {self.father is None and 'NULL' or self.father.indiv}\n"  # noqa
            f"Mother   : {self.mother is None and 'NULL' or self.mother.indiv}\n"  # noqa
            f"Children : [ {*self.children,} ]"
        )
        return output

    @property
    def gen(self):
        if self.__gen is None:
            self.__gen = self.compute_gen()
        return self.__gen

    @gen.setter
    def gen(self, v):
        if v is not None:
            raise ValueError("PedNode.gen can only be assigned None")
        self.__gen = None

    def compute_gen(self):
        if self.father is None and self.mother is None:
            return 0
        else:
            if self.father is None:
                return self.mother.gen + 1
            elif self.mother is None:
                return self.father.gen + 1
            else:
                return max(self.mother.gen + 1, self.father.gen + 1)

    @property
    def indiv(self):
        """Individual data"""
        return self.__indiv

    @indiv.setter
    def indiv(self, v):
        warnings.warn("[PedNode] : indiv attribute is read-only.")

    @property
    def father(self):
        """Father node"""
        return self.__father

    @father.setter
    def father(self, indiv):
        if indiv is None:
            self.__father = None
        elif self.__father is None:
            assert type(indiv) is PedNode
            self.__father = indiv
        else:
            raise ValueError("Setting a new father is not possible.")

    @property
    def mother(self):
        """Mother node"""
        return self.__mother

    @mother.setter
    def mother(self, indiv):
        if indiv is None:
            self.__mother = None
        elif self.__mother is None:
            assert type(indiv) is PedNode
            self.__mother = indiv
        else:
            raise ValueError("Setting a new mother is not possible.")

    @property
    def sex(self):
        """FEMALE or MALE or None"""
        return self.__sex

    @sex.setter
    def sex(self, v):
        if self.sex and v != self.sex:
            raise ValueError(f"Sex mismatch for individual {self.indiv}")
        else:
            assert v == MALE or v == FEMALE or v is None
            self.__sex = v

    def add_child(self, child):
        """Add a child to the node."""
        assert type(child) is PedNode
        if child not in self.children:
            self.children.append(child)


class Pedigree:
    """Pedigree relationships between individuals"""

    def __init__(self, members=[]):
        """
        Parameters
        ----------
        members : list of obj
            List of members to add to the pedigree at initialization.
            The type of objects in the list must be hashable.
        """
        self.nodes = {}
        self.__founders = None
        self.__non_founders = None
        self.__families = None
        for m in members:
            self.add_node(m)

    def __iter__(self):
        """Iterator on nodes in the pedigree.
        Parents are guaranteed to precede offspring.
        """
        for n in sorted(self.nodes.values(), key=lambda x: x.gen):
            yield n

    def to_tuples(self):
        """
        Stores relationships in the pedigree in tuples (indiv,father,mother)
        returns a list of such tuples
        """
        rels = []
        for node in self:
            fa = (node.father is None) and "0" or node.father.indiv
            mo = (node.mother is None) and "0" or node.mother.indiv
            rels.append((node.indiv, fa, mo))
        return rels

    @classmethod
    def from_tuples(cls, rels):
        ped = cls()
        for rel in rels:
            if rel[1] != "0":
                ped.set_father(rel[0], rel[1])
            if rel[2] != "0":
                ped.set_mother(rel[0], rel[2])
            if rel[1] == "0" and rel[2] == "0":
                ped.add_indiv(rel[0])
            _ = ped.nodes[rel[0]]
        return ped

    @classmethod
    def from_fam_file(cls, path, parent_from_FID=False, default_parent=MALE):
        with open(path) as f:
            relations = []
            for ligne in f:
                buf = ligne.split()
                ind = buf[1]
                if buf[2] == "0":
                    fa = None
                    if parent_from_FID and default_parent == MALE:
                        fa = buf[0]
                else:
                    fa = buf[2]
                if buf[3] == "0":
                    mo = None
                    if parent_from_FID and default_parent == FEMALE:
                        mo = buf[0]
                else:
                    mo = buf[3]
                relations.append((ind, fa, mo))
            ped = cls()
            for rel in relations:
                if rel[1] is not None:
                    ped.set_father(rel[0], rel[1])
                if rel[2] is not None:
                    ped.set_mother(rel[0], rel[2])
                if rel[1] is None and rel[2] is None:
                    ped.add_indiv(rel[0])
                _ = ped.nodes[rel[0]]
        return ped

    @classmethod
    def from_pednodes(cls, pednodes):
        """
        Create a new pedigree from a list of pednodes.

        Argument
        --------
        pednodes : list of PedNode objects

        """
        ped = cls()
        for n in pednodes:
            assert type(n) is PedNode
            ped.nodes[n.indiv] = n
        for n in ped.nodes.values():
            if n.father is not None:
                try:
                    _ = ped.nodes[n.father.indiv]
                except KeyError:
                    logger.warning(
                        "When creating new pedigree from pednodes : "
                        "father not found in nodes, deleting"
                    )
                    n.father = None
            if n.mother is not None:
                try:
                    _ = ped.nodes[n.mother.indiv]
                except KeyError:
                    warnings.warn(
                        "When creating new pedigree from pednodes : "
                        "mother not found in nodes, deleting"
                    )
                    n.mother = None
            putative_children = n.children[:]
            for c in putative_children:
                try:
                    _ = ped.nodes[c.indiv]
                except KeyError:
                    warnings.warn(
                        "When creating new pedigree from pednodes : "
                        "child not found, deleting"
                    )
                    n.children.remove(c)
        return ped

    @property
    def founders(self):
        """List of founders in the pedigree (no known parents)"""
        if self.__founders is None:
            self.__founders = [
                x for x in self.nodes.values() if x.father is None and x.mother is None
            ]
        return self.__founders

    @property
    def non_founders(self):
        """List of non founders in the pedigree (at least one know parent)"""
        if self.__non_founders is None:
            self.__non_founders = [
                x
                for x in self.nodes.values()
                if x.father is not None or x.mother is not None
            ]
        return self.__non_founders

    @property
    def families(self):
        """List of Pedigree objects. One for each independant family in the
        dataset"""
        if self.__families is None:
            self.__families = self.build_families()
        return self.__families

    def _reset_families(self):
        """Reset structural information in the pedigree"""
        # reset founders / non_founders / families information
        self.__founders = None
        self.__non_founders = None
        self.__families = None
        for n in self.nodes.values():
            n.gen = None

    def add_indiv(self, indiv):
        """Add a new guy to the pedigree"""
        if indiv in self.nodes:
            logger.debug(f"{indiv} is already there")
        else:
            self.nodes[indiv] = PedNode(indiv)

    def del_indiv(self, indiv):
        """Remove an individual from the pedigree. Links to its node are
        removed.

                Returns
                -------
                the node of the individual if found, else None.

        """
        try:
            n = self.nodes[indiv]
        except KeyError:
            return None
        else:
            del self.nodes[indiv]
            for c in n.children:
                if c.father and c.father == n:
                    c.father = None
                elif c.mother and c.mother == n:
                    c.mother = None
            if n.father:
                n.father.children.remove(n)
            if n.mother:
                n.mother.children.remove(n)
            return n

    def add_node(self, node):
        """Add a node in the pedigree.
        Familial links are restored.

        Argument:
        ---------
        node : PedNode object
            the node to add
        """
        assert type(node) == PedNode
        self.nodes[node.indiv] = node
        if node.father:
            node.father.children.append(node)
        if node.mother:
            node.mother.children.append(node)
        if node.sex:
            if node.sex == MALE:
                for c in node.children:
                    c.father = node
            elif node.sex == FEMALE:
                for c in node.children:
                    c.mother = node

    def set_father(self, offspring, dad):
        """Set that a dad -> offspring relationship

        If called more than once with a different dad a warning is
        raised and the first call takes precedence.  If dad and/or
        offspring are not in the pedigree they will be added.


        Parameters
        ----------
        offspring : object
            offspring to be set. Must be hashable.
        dad : object
           dad to be set. Must be hashable.
        """
        try:
            indiv = self.nodes[offspring]
        except KeyError:
            self.add_indiv(offspring)
            indiv = self.nodes[offspring]

        try:
            father = self.nodes[dad]
            if not father.sex:
                father.sex = MALE
        except KeyError:
            self.add_indiv(dad)
            father = self.nodes[dad]
            father.sex = MALE

        indiv.father = father
        father.add_child(indiv)
        self._reset_families()

    def get_father(self, indiv):
        try:
            member = self.nodes[indiv]
        except KeyError:
            raise KeyError("Individual not found in pedigree")
        return member.father

    def set_mother(self, offspring, mom):
        """Set that a mom -> offspring relationship

        If called more than once with a different mom a warning is
        raised and the first call takes precedence.  If mom and/or
        offspring are not in the pedigree they will be added.

        Parameters
        ----------
        offspring : object
           offspring to be set. Must be hashable.
        mom : object
           mom to be set. Must be hashable.
        """
        try:
            indiv = self.nodes[offspring]
        except KeyError:
            self.add_indiv(offspring)
            indiv = self.nodes[offspring]

        try:
            mother = self.nodes[mom]
            if not mother.sex:
                mother.sex = FEMALE
        except KeyError:
            self.add_indiv(mom)
            mother = self.nodes[mom]
            mother.sex = FEMALE

        indiv.mother = mother
        mother.add_child(indiv)
        self._reset_families()

    def get_mother(self, indiv):
        try:
            member = self.nodes[indiv]
        except KeyError:
            raise KeyError("Individual not found in pedigree")
        return member.mother

    def _get_relatives(self, node, connected=None):
        """Get all individuals connected to a node in the pedigree.

        Parameters
        ----------
        node : obj of class PedNode
           the investigated node
        connected : list of obj of class PedNode
           the list of ancestors and descendants of node, populated by
           calling this method.

        Returns
        -------
        list of obj
           list of PedNode objects, including node

        """
        if connected is None:
            connected = []
        if node in connected:
            return
        connected.append(node)
        for off in node.children:
            self._get_relatives(off, connected)
        if node.father:
            self._get_relatives(node.father, connected)
        if node.mother:
            self._get_relatives(node.mother, connected)
        return connected

    @property
    def unrelated_individuals(self):
        """List of unrelated individuals with no parents and no child in the
        pedigree"""
        return [x.indiv for x in self.founders if len(x.children) == 0]

    def del_unrelated(self):
        """Remove unrelated individuals from the pedigree"""
        torm = self.unrelated_individuals[:]
        for indiv in torm:
            logger.debug(f"Removing {indiv} as it is not related to anyone")
            self.del_indiv(indiv)

    def build_families(self):
        """
        Cluster individuals by families
        """
        founders2go = self.founders[:]
        offspring2go = self.non_founders[:]
        families = []
        while len(founders2go) > 0:
            relatives = self._get_relatives(founders2go[0])
            fam = Pedigree.from_pednodes(relatives)
            families.append(fam)
            for x in relatives:
                try:
                    founders2go.remove(x)
                except ValueError:
                    offspring2go.remove(x)
        return families

    def get_family(self, indiv):
        """Get subpedigree connected to an individual in the pedigree

        Argmuent
        ----------
        indiv: name of the focal individual

        Returns: object
        --------
        A pedigree object with all individuals connected to the focal
        individual in the pedigree

        """
        try:
            node = self.nodes[indiv]
        except KeyError:
            raise ValueError(f"Focal individual {indiv} not found in pedigree")
        sys.setrecursionlimit(len(self.nodes) + 1)
        relatives = self._get_relatives(node)
        fam = Pedigree.from_pednodes(relatives)
        return fam



import os
import logging
from collections import defaultdict
from multiprocessing import Pool
import numpy as np
from scipy.stats import binom
from cyvcf2 import VCF
from . import vcf, pedigree

logger = logging.getLogger(__name__)


def genotype_vector(genotypes):
    return np.array([vcf.geno2int(*g[:2]) for g in genotypes], dtype=int)


def mendel_errors(args):
    genotypes, pairs = args
    pairs = np.array(pairs)
    genotypes = np.array(genotypes)
    from_genotypes = np.array(genotypes)[pairs[:, 0], :2]
    to_genotypes = np.array(genotypes)[pairs[:, 1], :2]
    fg = genotype_vector(from_genotypes)
    tg = genotype_vector(to_genotypes)
    nobs = (fg > 0) & (tg > 0)
    nerr = ((fg == 0) & (tg == 2)) | ((fg == 2) & (tg == 0))
    return nerr, nobs


def identify_bad_pairs(pairs, merr, fpr=1e-3):
    """Identify outliers for mendelian errors.

    Starting from all pairs, outliers are identified iteratively as follows:

    1. Compute overal error rate
    2. For each pair compute the p-value (from Binomial) that data comes
       from the binomial
    3. reject the null is p-value < fpr/len(pairs) (Bonferroni correction)
    4. Remove pairs for which the null is rejected from the pair list

    These steps are iterated until no pair is removed from the set.

    Arguments
    ---------
    pairs : list of objects
        family links
    merr : array len(pairs) x 2
        array of number of errors (merr[:,0]) and number of
        informative loci(merr[:,1])
    fpr : float
        false positive rate to call outliers

    Returns
    -------
    list of objects
        List of pairs from the input list that are outliers
    """
    pair2rm = []
    pval_th = fpr / len(pairs)
    logger.info(f"Identifying links with p < {pval_th:.2g}")
    current_merr = merr.copy()
    current_pairs = pairs[:]
    while True:
        tx_err = np.sum(current_merr[:, 0]) / np.sum(current_merr[:, 1])
        logger.debug(f"te : {tx_err}")
        tmp_pairs = current_pairs[:]
        keepidx = np.ones(len(current_pairs), dtype=bool)
        for i, (p, e) in enumerate(zip(current_pairs, current_merr)):
            pval = binom.sf(n=e[1], p=tx_err, k=e[0])
            if pval < pval_th:
                logger.debug(f"{p} : pbinom({e[0]}, size={e[1]},  p={tx_err}) = {pval}")
                pair2rm.append(p)
                tmp_pairs.remove(p)
                keepidx[i] = False
        if keepidx.all():
            break
        logger.debug(f"Start with {len(current_pairs)} -> {len(tmp_pairs)}")
        current_pairs = tmp_pairs[:]
        current_merr = current_merr[keepidx, :]
    return pair2rm


def main(args):
    prfx = args.prfx
    vcf_file = f"{prfx}.vcf.gz"
    fam_file = f"{prfx}.fam"
    myvcf = VCF(vcf_file, gts012=True, strict_gt=True, lazy=True)
    ped = pedigree.Pedigree.from_fam_file(fam_file)

    # Identify indices
    indiv_idx = defaultdict(lambda: -1)
    for i, v in enumerate(myvcf.samples):
        try:
            node = ped.nodes[v]
            indiv_idx[node.indiv] = i
        except KeyError:
            continue

    # Find parent -> offspring pairs of indices
    pairs = []
    for node in ped:
        if indiv_idx[node.indiv] < 0:
            continue
        for c in node.children:
            if indiv_idx[c.indiv] < 0:
                continue
            pairs.append((indiv_idx[node.indiv], indiv_idx[c.indiv]))
    # pairs = np.array(pairs, dtype=int)

    geno_getter = ((s.genotypes, pairs) for s in myvcf)
    merr = np.zeros((len(pairs), 2), dtype=int)
    with Pool(args.c) as workers:
        for nerr, nobs in workers.imap_unordered(
            mendel_errors, geno_getter, chunksize=50
        ):
            merr[:, 0] += nerr
            merr[:, 1] += nobs
    tx_err = np.sum(merr[:, 0]) / np.sum(merr[:, 1])
    # pval_th = 0.001/pairs.shape[0]

    logger.info(f"Global Mendel Error rate : {tx_err:.2g}")
    # unset_links=[]
    unset_links = identify_bad_pairs(pairs, merr)
    with open(f"{prfx}.mendel.err", "w") as fout:
        print("parent offspring nerr nobs err.rate pvalue removed", file=fout)
        for p, e in zip(pairs, merr):
            pval = binom.sf(n=e[1], p=tx_err, k=e[0])
            print(
                f"{myvcf.samples[p[0]]} "
                f"{myvcf.samples[p[1]]} "
                f"{e[0]} {e[1]} {e[0]/e[1]:.2g} {pval:.2g} "
                f"{p in unset_links}",
                file=fout,
            )

    if len(unset_links) > 0:
        logger.info(f"Saving original fam file to {prfx}.fam.orig")
        os.replace(f"{prfx}.fam", f"{prfx}.fam.orig")
        logger.info("Correcting pedigree errors in fam file")
        with open(f"{prfx}.fam.orig") as fam:
            with open(f"{prfx}.fam.new", "w") as nfam:
                for ligne in fam:
                    buf = ligne.split()
                    try:
                        pat_p = (indiv_idx[buf[2]], indiv_idx[buf[1]])
                    except KeyError:
                        pass
                    else:
                        if pat_p in unset_links:
                            logger.info(f"removing paternal link {buf[2]} -> {buf[1]}")
                            buf[2] = "0"
                    try:
                        mat_p = (indiv_idx[buf[3]], indiv_idx[buf[1]])
                    except KeyError:
                        pass
                    else:
                        if mat_p in unset_links:
                            logger.info(f"removing maternal link {buf[3]} -> {buf[1]}")
                            buf[3] = "0"
                    print(" ".join(buf), file=nfam)
        logger.info(f"new fam file is : {prfx}.fam.new")



# -*-coding:utf-8 -*
"""
Yapp module recombination.py

Yapp module to analyse recombination information in pedigrees
"""
import sys
import logging
import collections.abc
from collections import defaultdict
import numpy as np
from scipy.interpolate import interp1d

from . import family_phaser, MALE, FEMALE

logger = logging.getLogger(__name__)


def reg2chr(bedstr):
    """Returns chromosome value from  bed string"""
    return bedstr.split(":")[0]


class RecombAnalyser:
    def __init__(self, phaser_db):
        self.phaser = family_phaser.Phaser.load(phaser_db)
        self.parents = {}
        for node in self.phaser.pedigree:
            # need at least 2 offspring to detect recomb
            if len(node.children) > 0:
                self.add_parent(node)
        # size_covered[sex][chrom]
        self.size_covered = defaultdict(lambda: defaultdict(lambda: 0.0))
        self._gsize = None

    @property
    def crossovers(self):
        """Iterator over crossing overs"""
        return (co for co in ind.meioses.values() for ind in self.parents)  # noqa

    @property
    def male_crossovers(self):
        """Iterator over male crossing overs"""
        return (
            co
            for co in ind.meioses.values()  # noqa
            for ind in self.parents
            if ind.sex == MALE
        )

    @property
    def female_crossovers(self):
        """Iterator over female crossing overs"""
        return (
            co
            for co in ind.meioses.values()  # noqa
            for ind in self.parents
            if ind.sex == FEMALE
        )

    def add_parent(self, pednode):
        try:
            par = self.parents[pednode.indiv]
        except KeyError:
            par = Parent(pednode.indiv, pednode.sex)
            self.parents[pednode.indiv] = par
        else:
            logger.warning(
                f"Trying to create the parent {pednode.indiv} multiple times"
            )

    @property
    def genome_size(self):
        res = 0
        for reg in self.phaser.regions:
            res += self.phaser.data["variants"][reg]["POS"][-1]
        return res

    @staticmethod
    def get_crossovers(si, call=0.999):
        """Find crossover boundaries from segregation indicators

        Arguments
        ---------
        - si: np.array( (2,L) )
           Segragation indicator. si[0,] is the best phase solution,
           si[1,] is the probability that the phase is correct. At
           each phase switch, determines the boundaries such that the
           phase probabilty is > call
        - call: float
           Minimum probability to consider a phase resolved.

        Returns
        -------

        list[ [l,r] ]
           list of crossover boundaries found (left, right) = Indices in
           the array.

        """
        best_guess = np.array([x[0] for x in si])
        phase_prob = np.array([x[1] for x in si])
        co_loc = np.asarray((best_guess[1:] - best_guess[:-1]) != 0).nonzero()[
            0
        ]  # noqa
        nco = len(co_loc)
        logger.debug("Found {nco} candidate crossovers")
        res = []
        if nco == 0:
            return res
        for ell in co_loc:
            left = ell
            while phase_prob[left] < call:
                if left == 0:
                    break
                left -= 1
            right = ell + 1
            while phase_prob[right] < call:
                if right == len(best_guess) - 1:
                    break
                right += 1
            if best_guess[left] != best_guess[right]:
                res.append((left, right))
        #  returns a set to resolve local clusters of crossovers
        return set(res)

    @staticmethod
    def min_interval_size(nmeio, recrate=1, alpha=0.05):
        """Determine minimal interval size to estimate recombination rates

        Assuming a Poisson distributed number of crossovers in an
        interval of size L.  The probability of sampling 0 crossovers
        among nmeio meioses is:

        p_0 = exp( - recrate * L * nmeio)

        The minimal interval size is the one that satisfies p_0 =
        alpha, or:

        L = -log(alpha)/(recrate*nmeio)

        Arguments
        ---------
        nmeio: int
          number of meioses
        recracte: float
          recombination rate (in cM/Mb)
        alpha: float
          precision parameter ( P(k=0)<alpha)

        Returns
        -------
        float
           minimum interval size

        """
        return 1e8 * np.log(alpha) / (recrate * nmeio)

    def run(self, recrate, call):
        """Run the recombination analysis"""
        logger.info("Finding recombinations")
        self.identify_crossovers(recrate=recrate, call=call)

    @staticmethod
    def get_chromosome_pair(genotype, gametes, segregations, seg_probs):
        si_pat = list(zip(segregations[0], seg_probs[0]))
        si_mat = list(zip(segregations[1], seg_probs[1]))
        chp = family_phaser.ChromosomePair.from_phase_data(
            genotype, gametes[0], gametes[1], si_pat, si_mat
        )
        return chp

    def set_informative_meioses(self):
        """Identify informative meioses for each parent"""
        logger.info("Set Informative meioses")
        # sample indices
        smpidx = {}
        for i, name in enumerate(self.phaser.data["samples"]):
            smpidx[name] = i
        for reg in self.phaser.regions:
            logger.info(f"Working on region {reg}")
            logger.info("Grabbing data")
            pos = self.phaser.data["variants"][reg]["POS"]
            mids = np.array([0.5 * (x + y) for x, y in zip(pos[:-1], pos[1:])])
            chrom = reg2chr(reg)
            genotypes = np.array(self.phaser.data["genotypes"][reg])
            gametes = np.array(self.phaser.data["phases"][reg]["gametes"])

            for indiv, par in self.parents.items():
                logger.info(f"Parent {indiv}")
                idx = smpidx[indiv]
                par_resolved = (genotypes[idx] == 1) & (
                    (gametes[idx][0] > 0) | (gametes[idx][1] > 0)
                )  # noqa
                node = self.phaser.pedigree.nodes[indiv]
                n_meio_info = np.zeros(len(pos) - 1, dtype=int)
                for c in node.children:
                    idx = smpidx[c.indiv]
                    par.meioses[c.indiv] = []
                    child_phased = (gametes[idx][0] > 0) | (gametes[idx][1] > 0)
                    combin_info = par_resolved & child_phased
                    infomk = combin_info.nonzero()[0]
                    if len(infomk) > 0:
                        infomk_l = min(infomk)
                        infomk_r = max(infomk)
                        n_meio_info[infomk_l:infomk_r] += 1
                        self.size_covered[node.sex][chrom] += (
                            pos[infomk_r] - pos[infomk_l]
                        ) * 1e-6
                    logger.debug(
                        f"\t Meiosis -> {c.indiv:24} "
                        f"{round( (pos[infomk_r]-pos[infomk_l])*1e-3)} Kb"
                    )
                par.set_n_info_meioses(chrom, mids, n_meio_info)
            for sex in self.size_covered:
                logger.info(
                    f"sex:{sex} "
                    f"chrom:{chrom} "
                    f"size:{self.size_covered[sex][chrom]} Mb"
                )

    def identify_crossovers(self, recrate=1, call=0.99):
        logger.info("Gathering crossovers")
        smpidx = {}
        for i, name in enumerate(self.phaser.data["samples"]):
            smpidx[name] = i

        for reg in self.phaser.regions:
            logger.info(f"Working on: {reg}")
            chrom = reg2chr(reg)

            pos = self.phaser.data["variants"][reg]["POS"]
            genotypes = np.array(self.phaser.data["genotypes"][reg])
            gametes = np.array(self.phaser.data["phases"][reg]["gametes"])
            segregations = np.array(self.phaser.data["phases"][reg]["segregations"])
            segprobs = np.array(self.phaser.data["phases"][reg]["seg_probs"])

            for node in self.phaser.pedigree:
                sys.stdout.write(f" --> {node.indiv:>24}\r")
                sys.stdout.flush()
                logger.debug(
                    f"{node.indiv} -- [ "
                    f"sex:{node.sex} "
                    f"gen:{node.gen} "
                    f"par:{(node.father!=None)+(node.mother!=None)} "
                    f"off:{len(node.children)} ]"
                )
                idx = smpidx[node.indiv]

                if node.father:
                    try:
                        par = self.parents[node.father.indiv]
                        idx_pat = smpidx[node.father.indiv]
                    except KeyError:
                        logger.debug(f"Cannot find {node.father.indiv}")
                        pass
                    else:
                        # crossovers
                        logger.debug("Finding paternal crossovers")
                        si_pat = list(
                            zip(
                                segregations[
                                    idx,
                                    0,
                                ],
                                segprobs[
                                    idx,
                                    0,
                                ],
                            )
                        )
                        cos = self.get_crossovers(si_pat, call=call)
                        for x, y in cos:
                            par.add_offspring_CO(node.indiv, chrom, pos[x], pos[y])
                        # coverage
                        par_resolved = (genotypes[idx_pat] == 1) & (
                            (gametes[idx_pat][0] > -1) & (gametes[idx_pat][1] > -1)
                        )  # noqa
                        child_phased = gametes[idx][0] > -1
                        combin_info = par_resolved & child_phased
                        infomk = combin_info.nonzero()[0]
                        if len(infomk) > 0:
                            infomk_l = min(infomk)
                            infomk_r = max(infomk)
                            par.add_offspring_coverage(
                                node.indiv, chrom, pos[infomk_l], pos[infomk_r]
                            )
                if node.mother:
                    try:
                        par = self.parents[node.mother.indiv]
                        idx_mat = smpidx[node.mother.indiv]
                    except KeyError:
                        pass
                    else:
                        si_mat = list(
                            zip(
                                segregations[
                                    idx,
                                    1,
                                ],
                                segprobs[
                                    idx,
                                    1,
                                ],
                            )
                        )
                        cos = self.get_crossovers(si_mat)
                        for x, y in cos:
                            par.add_offspring_CO(node.indiv, chrom, pos[x], pos[y])
                        # coverage
                        par_resolved = (genotypes[idx_mat] == 1) & (
                            (gametes[idx_mat][0] > -1) & (gametes[idx_mat][1] > -1)
                        )  # noqa
                        child_phased = gametes[idx][1] > -1  # noqa
                        combin_info = par_resolved & child_phased
                        infomk = combin_info.nonzero()[0]
                        if len(infomk) > 0:
                            infomk_l = min(infomk)
                            infomk_r = max(infomk)
                            par.add_offspring_coverage(
                                node.indiv, chrom, pos[infomk_l], pos[infomk_r]
                            )

    def write_results(self, prefix):
        with open(prefix + "_yapp_recomb_coverage.txt", "w") as fout:
            print("parent sex offspring chrom left right", file=fout)
            for name, par in self.parents.items():
                for off in par.coverage:
                    for chrom in par.coverage[off]:
                        print(
                            f"{name} "
                            f"{((par.sex==None) and 'U') or ((par.sex==MALE) and 'M' or 'F')} "  # noqa
                            f"{off} "
                            f"{chrom} "
                            f"{par.coverage[off][chrom][0]} "
                            f"{par.coverage[off][chrom][1]}",
                            file=fout,
                        )
        with open(prefix + "_yapp_recombinations.txt", "w") as fout:
            print("parent sex offspring chrom left right", file=fout)
            for name, par in self.parents.items():
                for off in par.meioses:
                    for co in sorted(par.meioses[off]):
                        print(
                            f"{name} "
                            f"{((par.sex==None) and 'U') or ((par.sex==MALE) and 'M' or 'F')} "  # noqa
                            f"{off} "
                            f"{co.chrom} {co.left} {co.right}",
                            file=fout,
                        )


class Parent:
    """Class for storing information for each parent

    Attributes
    ----------
    - name: str
        identifier for the parent
    - sex: int or None
        sex of the parent ( 0: Male, 1: Female)
    - meioses: dict( str: list of CrossingOver objects)
        meioses of the individual.
    - coverage: dict( str: dict of ( int, int))
        informative genome of each offspring(key).
        Values are [chromosome]=(left,right)
    - nmeio_info: function: x(chrom,pos) -> int
        a function that returns the number of meiosis at a given
        genomic coordinate
    """

    def __init__(self, name, sex=None):
        self.name = name
        self.sex = sex
        self.meioses = defaultdict(list)
        self.coverage = defaultdict(dict)
        self._nmeio_info = None

    @property
    def nmeioses(self):
        return len(self.meioses)

    @property
    def nb_CO_meioses(self):
        """
        List of the number of crossovers for each meiosis
        """
        return [len(v) for v in self.meioses.values()]

    @property
    def nb_CO_tot(self):
        """
        Total number of crossing over in all meioses
        """
        return np.sum(self.nb_CO_meioses)

    def get_offspring_CO(self, name):
        """
        Get the list of CO for offspring with name *name*
        """
        return self.meioses[name]

    def add_offspring_CO(self, name, chro, left, right):
        """
        Add a crossing over in offspring *name*
        on chromosome *chro* between *left* and *right*
        """
        myco = CrossingOver(chro, left, right)
        self.meioses[name].append(myco)

    def add_offspring_coverage(self, name, chro, left, right):
        """
        Add coveage for offspring *name*
        on chromosome chro between left and right
        """
        self.coverage[name][chro] = (left, right)

    def n_info_meioses(self, chrom, pos):
        """Get the number of informative meioses for the parent
        at position pos on chromosome chrom.

        Arguments
        ---------
        - chrom: int
            Chromosome
        - pos: int
            Position

        Returns
        -------
        int
           number of informative meioses
        """
        try:
            return self._nmeio_info(chrom, pos)
        except TypeError:
            return self.nmeioses

    def set_n_info_meioses(self, chrom, positions, values):
        """Enter information on the number of informative meioses on
        a chromosome, at a set of positions.

        Arguments
        ---------
        - chrom: int
            chromosome
        - positions: array of int
            Positions at which the number of informative meioses is known
        - values: array of int
            Number of informative meioses at each position
        """
        if self._nmeio_info is None:
            self._nmeio_info = Nmeioses(self.nmeioses)
        self._nmeio_info.add_predictor(chrom, positions, values)

    def n_info_meioses_seg(self, chrom, left, right):
        return max(self.n_info_meioses(chrom, left), self.n_info_meioses(chrom, right))

    def oi_xi(self, chrom, w_left, w_right):
        """Computes probabilities that each crossing over in the parent
        occurs in genomic region on chromosome *chrom* between
        positions *w_left* and *w_right*.

        Returns a tuple with entries:
        -- list of contributions for each CO
        -- number of informative meioses for the parent in the region

        """
        contrib = []
        for m in self.meioses.values():
            for co in m:
                if co.chro == chrom and not (
                    (w_right < co.left) or (w_left > co.right)
                ):  # noqa
                    contrib += [co.oi_xi(w_left, w_right)]
        return (contrib, self.n_info_meioses_seg(chrom, w_left, w_right))


class Nmeioses(collections.abc.Callable):
    """Class offering a callable interface to interpolate the number of
    informative meioses from observed data.

    Usage
    -----
    Nmeioses(chrom, pos) -> int

    Nmeioses.add_predictor(chrom, positions, values): set up a 1D interpolator
    for chromosome chrom from observed (positions , values) points.

    Returns
    -------
    int
       Interpolated number of meioses. 0 if outside training range

    """

    def __init__(self, default_value):
        self.default = int(default_value)
        self.predictors = defaultdict(lambda: lambda x: self.default)

    def __call__(self, chrom, pos):
        try:
            return int(np.ceil(self.predictors[chrom](pos)))
        except ValueError:
            return 0

    def add_predictor(self, chrom, positions, values):
        self.predictors[chrom] = interp1d(positions, values, fill_value=0)


class CrossingOver:
    """
    Class to store crossing over information

    Attributes:
    -- chro: chromosome
    -- left: position of the marker on the left side
    -- right: position of the marker on the right side
    """

    def __init__(self, chrom, left, right):
        assert right > left
        self.chrom = chrom
        self.left = left
        self.right = right

    def __lt__(self, other):
        return (self.chrom, self.left, self.right) < (
            other.chrom,
            other.left,
            other.right,
        )  # noqa

    @property
    def size(self):
        return self.right - self.left

    def oi_xi(self, w_left, w_right):
        """Computes the probability that the crossing over occured n the
        window between w_left and w_right

        """
        if (w_right < self.left) or (w_left > self.right):
            # no overlap
            # wl------wr               wl----------wr
            #              sl-------sr
            return 0
        elif (w_left <= self.left) and (self.right <= w_right):
            # co included in window
            # wl---------------------------wr
            #       sl---------------sr
            return 1
        elif (self.left <= w_left) and (w_right <= self.right):
            # window is included in co
            #       wl------wr
            # sl---------------------sr
            return float(w_right - w_left) / (self.right - self.left)
        elif w_left < self.left:
            # we know w_right<self.right as other case is treated above
            # wl-----------------wr
            #     sl------------------sr
            return float(w_right - self.left) / (self.right - self.left)
        else:
            # only case left
            #     wl-----------------wr
            # sl--------------sr
            try:
                assert (self.left <= w_left) and (self.right < w_right)
            except AssertionError:
                print(self.right, w_right, self.left, w_left)
                raise
            return float(self.right - w_left) / (self.right - self.left)


def main(args):
    prfx = args.prfx

    phaser_db = prfx + ".db"
    analyzer = RecombAnalyser(phaser_db)
    analyzer.run(recrate=args.rho, call=args.minsp)
    analyzer.write_results(prfx)



from setuptools import setup
import yapp


def readme():
    with open("README.md") as f:
        return f.read()


setup(
    name="yappgen",
    version=yapp.__version__,
    description="Yet Another Phasing Program",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://forgemia.inra.fr/bertrand.servin/yapp",
    author="Bertrand Servin",
    author_email="bertrand.servin@inrae.fr",
    license="LGPLv2.1",
    packages=["yapp"],
    python_requires="<3.10",
    install_requires=[
        "numpy >= 1.20.2",
        "scipy >= 1.6.2",
        "cyvcf2 >= 0.30.4",
        "zarr >= 2.7.0",
        "h5py >= 3.2.1",
        "fastphase >= 1.2",
        "numba >= 0.53.1",
        "pytoulbar2",
    ],
    scripts=["bin/fphtrain", "bin/yapp"],
    zip_safe=False,
)



""" Phasing Half-Sib families using the WCSP approach of Favier(2011)

Test data corresponds to phase information:

T1: 1 * * 0 1 * 1
T2: * * * 0 1 * 0
T3: 0 * * * 0 * 0

Informative pairs of markers:

T1: (0,3),(3,4),(4,6)
T2: (3,4),(4,6)
T3: (0,4),(4,6)

Genetic distance is assumed to be 0.1 cM between each marker pair

"""
import os, sys
from collections import defaultdict
import numpy as np
from pytoulbar2 import CFN

test_phase = {
    "T1": [1, -1, -1, 0, 1, -1, 1],
    "T2": [-1, -1, -1, 0, 1, -1, 0],
    "T3": [0, -1, -1, -1, 0, -1, 0],
    "T4": [-1, -1, -1, -1, -1, -1, -1],
}


class PhaseData:
    """
    Class to format phase information data into WCSP variables

    Parameters:
    -----------
     -- rawphase: dict of phase info, keys are offspring names, values
                  are phase vectors (0,1,-1)
     -- mkpos: list of marker positions in centiMorgans
               (optional, if not given assumes 0.01 cM between markers)

     Attributes:
     -----------
     -- rawphase: init data
     -- info_mk: numpy array of informative markers indices
     -- info_pairs: dict with pairs of informative markers
            markers as keys and [N+,N-] as values.

     TODO:
     -----
     -- genetic map

    """

    def __init__(self, rawphase, mkpos=None):
        self.rawphase = rawphase
        if mkpos is None:
            self.mkpos = None
            self.recombination = self.recrate
        else:
            self.mkpos = mkpos
            self.recombination = self.recrate_from_pos
        # info_pairs stores pairs of inform. markers as keys
        # and [N+,N-] as values.
        self.info_pairs = {}
        # For each offspring
        for k, v in self.rawphase.items():
            # get informative markers ...
            # infomk=[i for i,x in enumerate(v) if x>=0]
            v = np.array(v)
            infomk = np.where(v >= 0)[0]
            if len(infomk) < 2:
                continue
            # and corresponding successive pairs
            mypairs = zip(infomk[:-1], infomk[1:])
            # for each pair
            for p in mypairs:
                # get phase information
                npair = int(v[p[0]] == v[p[1]])  # + or - phase
                # adds it to the pool
                try:
                    Nvec = self.info_pairs[p]
                except KeyError:
                    self.info_pairs[p] = [0, 0]
                    Nvec = self.info_pairs[p]
                Nvec[1 - npair] += 1
        # informative markers are all mks seen in pairs
        info_mk = defaultdict(int)
        for k in self.info_pairs:
            info_mk[k[0]] += 1
            info_mk[k[1]] += 1
        self.info_mk = sorted(info_mk.keys())

    def recrate_from_pos(self, i, j):
        """Returns recombination rate between marker i and j
        from their positions (in cM).
        """
        dtot = abs(self.mkpos[i] - self.mkpos[j])
        return 0.5 * (1 - np.exp(-dtot / 50))

    @staticmethod
    def recrate(i, j, d=0.01):
        """
        returns rec. rate between marker indices i and j
        assuming each adj. markers are separeted by d (default 0.01) cM
        """
        dtot = abs(i - j) * d
        return 0.5 * (1 - np.exp(-dtot / 50))


class PhaseSolver:
    """
    A class to Solve the phase of a parent.

    Parameters:
    -----------

    -- mk_list : list of markers
    -- mk_pairs : dict of pairs of markers (see PhaseData class)
    -- recmap_f : function that takes as input a pair of markers
                  and returns their rec. rate.
    """

    def __init__(self, mk_list, mk_pairs, recmap_f):
        self.mk = list(mk_list)
        self.pairs = mk_pairs
        self.recf = recmap_f
        self.L = len(self.mk)
        try:
            assert self.L > 1
        except AssertionError as e:
            raise ValueError("WCSP problem must have size > 1") from e
        # Initialize Optim. model
        self.constraints = []

    def add_constraints(self):
        """Build constraints from marker pairs"""
        for p in sorted(self.pairs):
            Nkl = self.pairs[p]
            # gt mk indices
            k = self.mk.index(p[0])
            ell = self.mk.index(p[1])
            # get recomb rate
            rkl = self.recf(p[0], p[1])
            assert rkl > 0
            # get cost
            Wkl = (Nkl[0] - Nkl[1]) * np.log((1 - rkl) / rkl)
            # create constraints
            if Wkl < 0:
                cost = [-Wkl, 0.0, 0.0, -Wkl]
            else:
                cost = [0.0, Wkl, Wkl, 0.0]
            self.constraints.append(([k, ell], cost))

    def solve(self, verbose=-1, MAXCOST=None):
        """Solve the WCSP and returns the inferred gamete"""
        Problem = CFN(MAXCOST, resolution=4)
        # resolution=4 (number of digits after dot in floating costs)
        # vac=1 (add this option to perform VAC in preprocessing)
        # Creates an array of phase indicators
        for i in range(self.L):
            Problem.AddVariable("p" + str(i), range(2))
        # fix p[0] == 0
        Problem.CFN.wcsp.assign(0, 0)
        # add constraints
        for c in self.constraints:
            Problem.AddFunction(c[0], c[1])
        try:
            res = Problem.Solve()
        except Exception:
            if len(Problem.CFN.solution()) > 0:
                res = [Problem.CFN.solution()]
            else:
                res = None
                print("WCSP Solver failed")
        if res and len(res[0]) > 0:
            return res[0]
        else:
            return res
            print("We might crash")


if __name__ == "__main__":
    T = PhaseData(test_phase)
    print("Info pairs:", *T.info_pairs.items())
    print("Info mk:", T.info_mk)
    S = PhaseSolver(T.info_mk, T.info_pairs, T.recombination)
    S.add_constraints()
    phase = S.solve()
    print("Resolved phase:", *phase)
